#!/usr/bin/env python
from unittest import mock

from PySide6.QtCore import QPoint, QRect, Qt
from PySide6.QtGui import QColor, QKeySequence
from PySide6.QtTest import QTest
from PySide6.QtWidgets import QMessageBox

from asammdf.gui.serde import COLORS
from test.asammdf.gui.test_base import Pixmap
from test.asammdf.gui.widgets.test_BasePlotWidget import TestPlotWidget


class TestPlotShortcuts(TestPlotWidget):
    def setUp(self):
        """
        Events:
            - Open measurement file.
            - Create plot sub-window.
            - Ensure that signal isn't with dots.
            - Ensure that there is no grid displayed.
            - Ensure that there are no bookmarks shown.

        Evaluate
            - Evaluate that there is one active sub-window.
        Returns
        -------

        """
        super().setUp()
        # Open measurement file
        self.setUpFileWidget(measurement_file=self.measurement_file, default=True)
        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural sort")
        # Select channels -> Press PushButton "Create Window" -> "Plot"
        self.create_window(window_type="Plot")
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        # Preset for plot
        # Remove dots
        if self.plot.plot.with_dots:
            self.plot.plot.set_dots(False)
        # check if grid is available -> hide grid
        if not self.plot.hide_axes_btn.isFlat():
            QTest.mouseClick(self.plot.hide_axes_btn, Qt.MouseButton.LeftButton)
        # hide bookmarks if it's available
        if self.plot.show_bookmarks:
            self.plot.toggle_bookmarks(hide=True)
        # Ensure that plot is not` in Focused mode
        if self.plot.focused_mode:
            self.plot.toggle_focused_mode()
        # Ensure that plot cursor is not with circle
        if self.plot.plot.cursor1.show_circle:
            self.plot.plot.cursor1.show_circle = False

        # get shortcuts
        self.assertIsNotNone(self.load_shortcuts_from_json_file(self.plot))

        self.processEvents()

    def tearDown(self):
        # Ensure that at the end, button "No" is pressed for MessageBox question window
        with mock.patch("asammdf.gui.widgets.mdi_area.MessageBox.question") as mo_question:
            mo_question.return_value = QMessageBox.StandardButton.No
            super().tearDown()

    def test_statistics_shortcut(self):
        """
        Test Scope:
            To check if after pressing key "M", info about the selected channel is displayed properly,
                and after deleting all channels,
                buffer is clean and info about the last channel isn't displayed,

        Events:.
            - Select 3 signals and create a plot.
            - Press Key "M".
            - Select First channel.
            - Pres key "Down".
            - Delete all channels.

        Evaluate:
            - Evaluate that Statistics isn't visible by default.
            - Evaluate that Statistics is visible after pressing key "M".
            - Evaluate that displayed info is related to the first channel.
            - Evaluate that displayed info is related to the second channel.
            - Evaluate that buffer is clear, so the displayed message is the default one.
        """
        # Setup
        default_message = "Please select a single channel"
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35, 36, 37]))
        # Evaluate if info isn't visible before pressing key M
        self.assertFalse(self.plot.info.isVisible())

        # Event
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["statistics"]))
        # Evaluate if info is visible after pressing key M
        self.assertTrue(self.plot.info.isVisible())

        # click on a first channel
        self.mouseClick_WidgetItem(self.channels[0])
        # Evaluate
        with self.subTest("test_key_M_click_on_first_channel"):
            self.assertEqual(self.plot.info._name, self.channels[0].name)
            self.assertEqual(self.plot.info.color, self.channels[0].color.name())

        # Press key "Down"
        QTest.keySequence(self.plot.channel_selection, QKeySequence("Down"))
        # Evaluate
        with self.subTest("test_key_M_press_key_Down_on_channel_selection"):
            self.assertEqual(self.plot.info._name, self.channels[1].name)
            self.assertEqual(self.plot.info.color, self.channels[1].color.name())

        # delete all channels
        QTest.keySequence(self.plot, QKeySequence("Ctrl+A"))
        QTest.keySequence(self.plot.channel_selection, QKeySequence("Delete"))
        # Evaluate
        with self.subTest("test_key_M_delete_all_channels"):
            # The default message is displayed instead of channel name
            self.assertNotEqual(self.plot.info._name, default_message)

    def test_focused_mode_shortcut(self):
        """
        Test Scope:
            To check if is displayed only selected channel after pressing key "2"
        Events:
            - Select 3 signals and create a plot
            - Press Key "2"
            - Select first channel
            - Pres key "Down"
            - Select third channel
        Evaluate:
            - Evaluate that three signals are available.
            - Evaluate that plot is not black and contains colors of all 3 channels.
            - Evaluate that plot is black after pressing key "2".
            - Evaluate that plot contains only color of first channel after clicking on first channel.
            - Evaluate that plot contains only color of second channel after pressing key "Down".
            - Evaluate that plot contains only color of third channel after clicked on third channel.
            - Evaluate that plot contains colors of all 3 channels after pressing key "2".
        """
        # Setup
        if not self.plot.focused_mode_btn.isFlat():
            QTest.mouseClick(self.plot.focused_mode_btn, Qt.MouseButton.LeftButton)
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35, 36, 37]))
        pixmap = self.plot.plot.viewport().grab()
        # Evaluate
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[1].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[2].color.name()))

        # case 0
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["focused_mode"]))
        self.processEvents(1)
        for _ in range(100):
            self.processEvents(0.01)
        # Evaluate
        self.assertTrue(Pixmap.is_black(self.plot.plot.viewport().grab()))

        # case 1
        self.mouseClick_WidgetItem(self.channels[0])
        for _ in range(50):
            self.avoid_blinking_issue(self.plot.channel_selection)
        pixmap = self.plot.plot.viewport().grab()
        # Evaluate
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[1].color.name()))
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[2].color.name()))

        # case 2
        QTest.keyClick(self.plot.channel_selection, Qt.Key_Down)
        for _ in range(50):
            self.avoid_blinking_issue(self.plot.channel_selection)
        pixmap = self.plot.plot.viewport().grab()
        # Evaluate
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[1].color.name()))
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[2].color.name()))

        # case 3
        self.mouseClick_WidgetItem(self.channels[2])
        for _ in range(50):
            self.avoid_blinking_issue(self.plot.channel_selection)
        pixmap = self.plot.plot.viewport().grab()
        # Evaluate
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[1].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[2].color.name()))

        # case 4
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["focused_mode"]))
        self.processEvents(0.1)
        pixmap = self.plot.plot.viewport().grab()
        # Evaluate
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[1].color.name()))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[2].color.name()))

    def test_ascii__bin__hex__physical_shortcuts(self):
        """
        Test Scope:
            Check if values are converted to hex, bin, phys and ascii after pressing the combination of key
                "Ctrl+<H>|<B>|<P>|<T>"

        Events:
            - Display 1 signal on plot.
            - Press "Ctrl+B".
            - Press "Ctrl+H".
            - Press "Ctrl+P".
            - Press "Ctrl+T".

        Evaluate:
            - Evaluate that unit is changed to hex after pressing key "Ctrl+H".
            - Evaluate that unit is changed to bin after pressing key "Ctrl+B".
            - Evaluate that unit is changed to phys (int value) after pressing key "Ctrl+P".
            - Evaluate that unit is changed to ascii after pressing key "Ctrl+T".
        """
        # add channels to plot and select it
        self.assertIsNotNone(self.add_channels([35]))
        self.mouseClick_WidgetItem(self.channels[0])

        # Store text from the selected channel
        physical = self.plot.selected_channel_value.text()
        physical_hours = int(physical.split(" ")[0])
        # Press "Ctrl+B"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["bin"]))
        bin_ = self.plot.selected_channel_value.text()
        bin_hours = int(bin_.split(" ")[0].replace(".", ""), 2)

        # Evaluate
        self.assertNotEqual(physical, bin_)
        self.assertEqual(physical_hours, bin_hours)

        # Press "Ctrl+H"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["hex"]))
        hex_ = self.plot.selected_channel_value.text()
        # Convert hex value to Int
        hex_hours = int(hex_.split(" ")[0], 16)

        # Evaluate
        self.assertNotEqual(physical, hex_)
        self.assertIn("0x", hex_)
        self.assertEqual(physical_hours, hex_hours)

        # Press "Ctrl+P"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["physical"]))
        new_physical = self.plot.selected_channel_value.text()
        new_physical_hours = int(new_physical.split(" ")[0])

        # Evaluate
        self.assertEqual(physical, new_physical)
        self.assertEqual(physical_hours, new_physical_hours)

        # Press "Ctrl+P"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["ascii"]))
        ascii_value = self.plot.selected_channel_value.text()
        ascii_int = ascii(ascii_value.split(" ")[0])
        ascii_int = ascii_int.split("'")[1]
        ascii_int = "0" + ascii_int.split("\\")[1]
        ascii_int = int(ascii_int, 16)

        # Evaluate
        self.assertNotEqual(physical, ascii_value)
        self.assertEqual(physical_hours, ascii_int)

    def test_raw__scaled_samples_shortcuts(self):
        """
        Test Scope:
            Check functionality of key "Alt+R" and "Alt+S". They must convert samples to raw and scaled forms.
        Events:
            - Display 1 signal on plot
            - Press "Alt+R"
            - Press "Alt+S"
        Evaluate:
            - Evaluate that signal mode is raw and line style is DashLine after pressing key "Alt+R"
            - Evaluate that signal mode is phys and line style is SolidLine after pressing key "Alt+S"
        """
        # add channels to plot
        self.assertIsNotNone(self.add_channels(["ASAM.M.SCALAR.SBYTE.LINEAR_MUL_2"]))
        phys_value = self.plot.selected_channel_value.text().split()[0]
        expected_raw_value = float(phys_value) / self.channels[0].signal.conversion.a
        # Press "Alt+R"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["raw_samples"]))
        self.processEvents(1)
        # Evaluate
        # Signal mode = raw
        self.assertEqual(self.channels[0].mode, "raw")
        # Signal line style = Dash line
        self.assertEqual(self.plot.plot.signals[0].pen.style(), Qt.PenStyle.DashLine)
        # Raw value
        self.assertEqual(expected_raw_value, float(self.plot.selected_channel_value.text()))

        # Press "Alt+S"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["scaled_samples"]))
        self.processEvents(1)
        # Evaluate
        # Signal mode = raw
        self.assertEqual(self.channels[0].mode, "phys")
        # Signal line style = Dash line
        self.assertEqual(self.plot.plot.signals[0].pen.style(), Qt.PenStyle.SolidLine)

    def test_insert_bookmark_shortcut(self):
        """
        Test Scope:
            Check if bookmark is created after pressing key "Ctrl+I"
        Events:
            - Open 'FileWidget' with valid measurement.
            - Display 1 signal on plot
            - Press "Ctrl+I" and hit "Ok"
        Evaluate:
            - Evaluate that bookmarks are not displayed before pressing "Ctrl+I"
            - Evaluate that bookmarks are displayed after pressing "Ctrl+I"
                and the message of the last bookmark is the first element of the returned list of mock object
        """
        QTest.mouseClick(
            self.plot.plot.viewport(), Qt.MouseButton.LeftButton, pos=self.plot.plot.viewport().geometry().center()
        )
        self.processEvents()
        timestamp = self.plot.plot.cursor1.value()
        # mock for bookmark
        with mock.patch("asammdf.gui.widgets.plot.QtWidgets.QInputDialog.getMultiLineText") as mo_getMultiLineText:
            mo_getMultiLineText.return_value = [self.id(), True]
            # Press "Ctrl+I"
            QTest.keySequence(self.plot, QKeySequence(self.shortcuts["insert_bookmark"]))

        mo_getMultiLineText.assert_called()

        # Destroy current widget
        with mock.patch("asammdf.gui.widgets.mdi_area.MessageBox.question") as mo_question:
            mo_question.return_value = QMessageBox.StandardButton.Yes
            self.destroy(self.widget)

        # Open file with new bookmark in new File Widget
        self.setUpFileWidget(measurement_file=self.measurement_file, default=True)
        self.processEvents(0.1)
        self.widget.channel_view.setCurrentText("Natural sort")
        # Select channels -> Press PushButton "Create Window" -> "Plot"
        self.create_window(window_type="Plot")
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()
        self.processEvents(0.1)

        pg_colors = Pixmap.color_names_exclude_defaults(self.plot.plot.viewport().grab())
        bookmarks_colors = COLORS[: len(COLORS) - len(self.plot.plot.bookmarks) - 1 : -1]

        # Evaluate
        self.assertTrue(self.plot.show_bookmarks)
        self.assertEqual(self.plot.plot.bookmarks[len(self.plot.plot.bookmarks) - 1].message, self.id())
        self.assertEqual(self.plot.plot.bookmarks[len(self.plot.plot.bookmarks) - 1].value(), timestamp)
        for bookmark, color in zip(self.plot.plot.bookmarks, bookmarks_colors, strict=False):
            self.assertEqual(bookmark.color, color)
            self.assertIn(color, pg_colors)

    def test_toggle_bookmarks_shortcut(self):
        """
        Test Scope:
            Check functionality of key "Ctrl+I". It must toggle bookmarks visibility on plot.
        Events:
            - Display 1 signal on plot
            - Press 3 times "Alt+I"
        Evaluate:
            - Evaluate that bookmarks are not displayed before pressing "Alt+I"
            - Evaluate that bookmarks are displayed after pressing "Alt+I" first time
            - Evaluate that bookmarks are not displayed after pressing "Alt+I" second time
            - Evaluate that bookmarks are displayed after pressing "Alt+I" third time
        """
        # Evaluate
        self.assertFalse(self.plot.show_bookmarks)
        # Press "Al+I"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["toggle_bookmarks"]))
        # Evaluate
        self.assertTrue(self.plot.show_bookmarks)
        # Press "Al+I"
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["toggle_bookmarks"]))
        # Evaluate
        self.assertFalse(self.plot.show_bookmarks)

    def test_edit_y_axis_scaling_shortcut(self):
        """
        Test Scope:
            Check if signal is changed his Y limits by setting it with ScaleDialog object called with "Ctrl+G"
        Events:
            - Create a plot widget with signals
            - Click on first channel
            - Press Ctrl+G
            - Set scaling and offset values
            - Hit "Ok"
        Evaluate:
            - Evaluate that Signal is on top and bottom part of plot
            - Evaluate that mock object was called
            - Evaluate that after pressing Ctrl+G, selected channel is situated on top and not on bottom part of plot
            - Evaluate if y_range of signal is identical with expected range
        """
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35]))
        self.processEvents()

        # Setup
        offset = 50.0
        scale = self.plot.plot.y_axis.range[1] - self.plot.plot.y_axis.range[0]
        y_bottom = -offset * scale / 100
        y_top = y_bottom + scale

        expected_y_range = y_bottom, y_top

        # Top
        pixmap = self.plot.plot.grab(QRect(0, 0, self.plot.plot.width(), int(self.plot.plot.height() / 2)))
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        # Bottom
        pixmap = self.plot.plot.grab(
            QRect(0, int(self.plot.plot.height() / 2), self.plot.plot.width(), int(self.plot.plot.height() / 2))
        )
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))

        # Click on the first channel
        self.mouseClick_WidgetItem(self.channels[0])
        with mock.patch("asammdf.gui.widgets.plot.ScaleDialog") as mo_ScaleDialog:
            mo_ScaleDialog.return_value.offset.value.return_value = offset
            mo_ScaleDialog.return_value.scaling.value.return_value = scale
            # Press Ctrl+Shift+C
            QTest.keySequence(self.plot, QKeySequence(self.shortcuts["edit_y_axis_scaling"]))

        # Evaluate

        # Evaluate that ScaleDialog object was created, and it's exec() method was called
        mo_ScaleDialog.return_value.exec.assert_called()
        self.processEvents(1)

        # Evaluate y_range tuple
        self.assertTupleEqual(self.plot.plot.signals[0].y_range, expected_y_range)

        # Top
        pixmap = self.plot.plot.grab(QRect(0, 0, self.plot.plot.width(), int(self.plot.plot.height() / 2)))
        # Evaluate plot, top half of plot must contain channel color
        self.assertTrue(Pixmap.has_color(pixmap, self.channels[0].color.name()))
        # Bottom
        pixmap = self.plot.plot.grab(
            QRect(0, int(self.plot.plot.height() / 2 + 2), self.plot.plot.width(), int(self.plot.plot.height() / 2 - 2))
        )
        # Evaluate plot, bottom half of plot must not contain channel color
        self.assertFalse(Pixmap.has_color(pixmap, self.channels[0].color.name()))

    def test_color_shortcut(self):
        """
        Test Scope:
            - Ensure that channel color is changed.

        Events:
            - Open Plot with two channels.
            - Press C.
            - Select one channel -> press C.
            - Select all channels -> press C.

        Evaluate:
            - Evaluate that color dialog is not open if the channel is not selected.
            - Evaluate that channel color is changed only for the selected channel.
        """
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35, 36, 37]))
        with self.subTest("test_WOSelectedChannel"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                QTest.keySequence(self.plot, QKeySequence(self.shortcuts["set_color"]))
                mo_getColor.assert_not_called()

        with self.subTest("test_1SelectedChannel"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                self.mouseClick_WidgetItem(self.channels[1])
                previous_color = self.channels[1].color.name()
                color = QColor("magenta")
                mo_getColor.return_value = color
                # Event
                QTest.keySequence(self.plot, QKeySequence(self.shortcuts["set_color"]))
                # Evaluate
                mo_getColor.assert_called()
                self.assertNotEqual(previous_color, self.channels[1].color.name())
                self.assertEqual(color.name(), self.channels[1].color.name())

        with self.subTest("test_AllSelectedChannel"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                color = QColor("black")
                mo_getColor.return_value = color
                # Set selected both channels
                QTest.keySequence(self.plot.channel_selection, QKeySequence("Ctrl+A"))
                # store previous colors of channels
                previous_ch_35_color = self.channels[0].color.name()
                previous_ch_36_color = self.channels[1].color.name()
                previous_ch_37_color = self.channels[2].color.name()
                color = QColor("black")
                mo_getColor.return_value = color
                # Event
                QTest.keySequence(self.plot, QKeySequence(self.shortcuts["set_color"]))
                # Evaluate
                mo_getColor.assert_called()

                self.assertNotEqual(previous_ch_35_color, self.channels[0].color.name())
                self.assertNotEqual(previous_ch_36_color, self.channels[1].color.name())
                self.assertNotEqual(previous_ch_37_color, self.channels[2].color.name())
                self.assertEqual(color.name(), self.channels[0].color.name())
                self.assertEqual(color.name(), self.channels[1].color.name())
                self.assertEqual(color.name(), self.channels[2].color.name())

    def test_copy__paste_channel_structure_shortcut(self):
        """
        Test Scope:
            - Ensure that selected channel is copied to clipboard and pasted into a plot.
        Events:
            - Open Plot with 2 channels
            - Select first channel
            - Press Ctrl+C
            - Press Ctrl+V
            - Open new plot window
            - Press Ctrl+V
        Evaluate:
            - Evaluate that is one more channel in channel selection area with the same properties as selected channel
            - Evaluate that channel is inserted in new window with the same properties
        """
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35, 36]))
        self.mouseClick_WidgetItem(self.channels[0])
        # Press Ctrl+C -> Ctrl+V
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["copy_channel_structure"]))
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["paste_channel_structure"]))
        self.processEvents()
        # Evaluate that now is three channels available
        self.assertEqual(3, self.plot.channel_selection.topLevelItemCount())
        original_name = self.channels[0].name
        original_color_name = self.channels[0].color.name()
        original_origin_uuid = self.channels[0].origin_uuid
        original_signal = self.channels[0].signal
        replica = self.plot.channel_selection.topLevelItem(1)
        # Evaluate channels essential attributes
        self.assertEqual(original_name, replica.name)
        self.assertEqual(original_color_name, replica.color.name())
        self.assertEqual(original_origin_uuid, replica.origin_uuid)
        self.assertEqual(original_signal, replica.signal)

        # Uncheck channels
        ch = self.find_channel(self.widget.channels_tree, self.channels[0].name)
        ch.setCheckState(0, Qt.CheckState.Unchecked)
        ch = self.find_channel(self.widget.channels_tree, self.channels[1].name)
        ch.setCheckState(0, Qt.CheckState.Unchecked)
        # Press PushButton "Create Window" -> "Plot"
        self.create_window(window_type="Plot")
        # Evaluate that second window is created
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 2)
        # Second window
        self.plot = self.widget.mdi_area.subWindowList()[1].widget()
        # Press Ctrl+V
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["paste_channel_structure"]))
        self.processEvents()
        # Evaluate that now is three channels available
        self.assertEqual(1, self.plot.channel_selection.topLevelItemCount())

        replica = self.plot.channel_selection.topLevelItem(0)
        # Evaluate channels essential attributes
        self.assertEqual(original_name, replica.name)
        self.assertEqual(original_color_name, replica.color.name())
        self.assertEqual(original_origin_uuid, replica.origin_uuid)
        self.assertEqual(original_signal, replica.signal)

    def test_copy__paste_display_properties_shortcuts(self):
        """
        Test Scope:
            Check if only display properties of selected channels is copied on another channel
                by shortcuts "Ctrl+ShiftC"->"Ctrl+Shift+V"
        Events:
            - Select 2 signals and create a plot
            - Click on first channel
            - Press key Ctrl+Shift+C
            - Click on second channel
            - Press key Ctrl+Shift+V
        Evaluate:
            - Evaluate that names and colors are different for both channels
            - Evaluate that after pressing shortcuts combination, names are different,
                but colors are the same for both channels
        """
        # add channels to plot
        self.assertIsNotNone(self.add_channels([35, 36, 37]))
        # Evaluate precondition
        self.assertNotEqual(self.channels[1].name, self.channels[2].name)
        self.assertNotEqual(self.channels[1].color.name(), self.channels[2].color.name())

        # Click on channel 36
        self.mouseClick_WidgetItem(self.channels[1])
        # Press Ctrl+Shift+C
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["copy_display_properties"]))
        self.mouseClick_WidgetItem(self.channels[2])
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["paste_display_properties"]))

        # Evaluate
        self.assertNotEqual(self.channels[1].name, self.channels[2].name)
        self.assertEqual(self.channels[1].color.name(), self.channels[2].color.name())
        self.assertEqual(self.channels[1].precision, self.channels[2].precision)
        self.assertEqual(self.channels[1].format, self.channels[2].format)
        self.assertEqual(self.channels[1].signal.y_link, self.channels[2].signal.y_link)
        self.assertEqual(self.channels[1].signal.individual_axis, self.channels[2].signal.individual_axis)
        self.assertListEqual(self.channels[1].ranges, self.channels[2].ranges)
        self.assertEqual(self.channels[1].signal.y_range[0], self.channels[2].signal.y_range[0])
        self.assertEqual(self.channels[1].signal.y_range[1], self.channels[2].signal.y_range[1])

    def test_increase__decrease_font_shortcuts(self):
        """
        Test scope:
            Ensure that Ctrl+[ and Ctrl+] will change font size

        Events:
            - Press Ctrl+[
            - Press Ctrl+]

        Evaluate:
         - Evaluate that font size was increased after shortcut Ctrl+[ was pressed
         - Evaluate that font size was decreased after shortcut Ctrl+] was pressed
        """
        font_size = self.plot.font().pointSize()
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["decrease_font"]))
        self.assertLess(font_size, self.plot.font().pointSize())

        font_size = self.plot.font().pointSize()
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["increase_font"]))
        self.assertGreater(font_size, self.plot.font().pointSize())

    def test_navigate_trough_zoom_history_shortcuts(self):
        """
        Test Scope:
            Test functionality of keys Backspace, Shift+Backspace, Shift+W.
            Those keys must change zoom history

        Events:
            - Add one channel to plot
            - Perform some zoom actions
            - Press key Backspace
            - Press Shift+backspace
            - Press Shift+W

        Evaluate:
            - Evaluate that after pressing keys "Shift+W", y_range is equal with y_range at start
            - Evaluate that after pressing keys "Shift+Backspace", y_range is equal with y_range after zoom action
            - Evaluate that after pressing key "Backspace", y_range isn't equal with previous ranges
        """
        # Setup
        if not self.plot.focused_mode_btn.isFlat():
            QTest.mouseClick(self.plot.focused_mode_btn, Qt.MouseButton.LeftButton)
        self.plot.plot.viewbox.menu.set_x_zoom_mode()
        self.plot.plot.viewbox.menu.set_y_zoom_mode()
        self.plot.plot.viewbox.setMouseMode(self.plot.plot.viewbox.PanMode)
        self.plot.plot.setFocus()
        self.processEvents(0.1)
        # add channel to plot
        self.assertIsNotNone(self.add_channels([35]))

        x = round(self.plot.plot.width() / 2)
        y = round(self.plot.plot.height() / 2)
        # Y range at start
        y_range = self.plot.plot.signals[0].y_range

        # Events
        self.mouseClick_WidgetItem(self.channels[0])
        self.avoid_blinking_issue(self.plot.channel_selection)
        # Click in the center of the plot
        QTest.mouseClick(
            self.plot.plot.viewport(),
            Qt.MouseButton.LeftButton,
            Qt.KeyboardModifier.NoModifier,
            QPoint(int(x), int(y)),
        )
        self.processEvents(0.01)
        # Rotate mouse wheel
        self.wheel_action(self.plot.plot.viewport(), x, y, 2)
        self.processEvents(0.01)

        # Click on plot
        QTest.mouseClick(
            self.plot.plot.viewport(),
            Qt.MouseButton.LeftButton,
            Qt.KeyboardModifier.NoModifier,
            QPoint(int(x * 0.5), int(y * 0.8)),
        )
        self.processEvents(0.01)

        # Rotate mouse wheel
        self.wheel_action(self.plot.plot.viewport(), x, y, -3)
        self.processEvents(1)
        self.processEvents(0.1)
        new_y_range = self.plot.plot.signals[0].y_range

        # Evaluate
        self.assertNotEqual(new_y_range, y_range)

        # click on Backspace
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["undo_zoom"]))
        self.processEvents()
        undo_zoom_y_range = self.plot.plot.signals[0].y_range

        # click on Shift + Backspace
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["redo_zoom"]))
        self.processEvents()
        redo_zoom_y_range = self.plot.plot.signals[0].y_range

        # click on Shift + W
        QTest.keySequence(self.plot, QKeySequence(self.shortcuts["x_default_zoom"]))
        self.processEvents()
        default_zoom_y_range = self.plot.plot.signals[0].y_range

        # Evaluate
        # Shift+W
        self.assertEqual(y_range[0], default_zoom_y_range[0])
        self.assertEqual(y_range[1], default_zoom_y_range[1])
        # Shift + Backspace
        self.assertEqual(new_y_range[0], redo_zoom_y_range[0])
        self.assertEqual(new_y_range[1], redo_zoom_y_range[1])
        # Backspace
        self.assertNotIn(undo_zoom_y_range[0], [y_range[0], new_y_range[0]])
        self.assertNotIn(undo_zoom_y_range[1], [y_range[1], new_y_range[1]])



"""
module to generate benchmark graphs from textul result file
"""

from typing import Literal

import matplotlib.pyplot as plt
import numpy as np


def generate_graphs(result: str, topic: str, aspect: Literal["ram", "time"], for_doc: bool = False) -> None:
    """genrate graphs from result file

    Parameters
    ----------
    result : str
        path to result file
    topic : str
        benchmark topic; for example "Open file" or "Save file"
    aspect : str
        performance indiitemsor; can be "ram" (RAM memory usage) or "time" (elapsed time)
    for_doc : bool
        wether the source code is used inside the documentation

    """
    with open(result) as f:
        lines = f.readlines()

    platform = "x86" if "32 bit" in lines[2] else "x64"

    idx = [i for i, line in enumerate(lines) if line.startswith("==")]

    table_spans = {
        "open": [idx[1] + 1, idx[2]],
        "save": [idx[4] + 1, idx[5]],
        "get": [idx[7] + 1, idx[8]],
        "convert": [idx[10] + 1, idx[11]],
        "merge": [idx[13] + 1, idx[14]],
    }

    start, stop = table_spans[topic.lower()]

    items = [l[:50].strip(" \t\r\n\0*") for l in lines[start:stop]]
    time = np.array([int(l[50:61].strip(" \t\r\n\0*")) for l in lines[start:stop]])
    ram = np.array([int(l[61:].strip(" \t\r\n\0*")) for l in lines[start:stop]])

    if aspect == "ram":
        array = ram
    else:
        array = time

    y_pos = list(range(len(items)))

    fig, ax = plt.subplots()
    fig.set_size_inches(15, 3.8 / 12 * len(items) + 1.2)

    asam_pos = [i for i, c in enumerate(items) if c.startswith("asam")]
    mdfreader_pos = [i for i, c in enumerate(items) if c.startswith("mdfreader")]

    ax.barh(asam_pos, array[asam_pos], color="green", ecolor="green")
    ax.barh(mdfreader_pos, array[mdfreader_pos], color="blue", ecolor="black")
    ax.set_yticks(y_pos)
    ax.set_yticklabels(items)
    ax.invert_yaxis()
    ax.set_xlabel("Time [ms]" if aspect == "time" else "RAM [MB]")
    if topic == "Get":
        ax.set_title("Get all channels (36424 calls) - {}".format("time" if aspect == "time" else "ram usage"))
    else:
        ax.set_title("{} test file - {}".format(topic, "time" if aspect == "time" else "ram usage"))
    ax.xaxis.grid()

    fig.subplots_adjust(
        bottom=0.72 / fig.get_figheight(),
        top=1 - 0.48 / fig.get_figheight(),
        left=0.4,
        right=0.9,
    )

    if aspect == "time":
        if topic == "Get":
            name = f"{platform}_get_all_channels.png"
        else:
            name = f"{platform}_{topic.lower()}.png"
    else:
        if topic == "Get":
            name = f"{platform}_get_all_channels_ram_usage.png"
        else:
            name = f"{platform}_{topic.lower()}_ram_usage.png"

    if for_doc:
        plt.show()
    else:
        plt.savefig(name, dpi=300)


files = (
    r"e:\02__PythonWorkspace\asammdf\benchmarks\results\x64_asammdf_2.7.0_mdfreader_0.2.7.txt",
    r"e:\02__PythonWorkspace\asammdf\benchmarks\results\x86_asammdf_2.7.0_mdfreader_0.2.7.txt",
)

for file in files:
    for topic in ("Open", "Save", "Get", "Convert", "Merge"):
        for aspect in ("time", "ram"):
            generate_graphs(file, topic, aspect)



import bisect
from functools import partial
import json
import os
from pathlib import Path
import re
from threading import Lock
from traceback import format_exc

from natsort import natsorted
import numpy as np
from numpy import searchsorted
import pyqtgraph.functions as fn
from PySide6 import QtCore, QtGui, QtWidgets

from asammdf.gui import utils
from asammdf.gui.dialogs.range_editor import RangeEditor
from asammdf.gui.utils import (
    copy_ranges,
    get_colors_using_ranges,
    unique_ranges,
    value_as_str,
)
from asammdf.gui.widgets.plot import PlotSignal
import asammdf.mdf as mdf_module

from .. import serde
from ..serde import ExtendedJsonDecoder, ExtendedJsonEncoder, extract_mime_names
from ..ui.numeric_offline import Ui_NumericDisplay
from ..utils import FONT_SIZE
from .tree import substitude_mime_uuids

HERE = Path(__file__).resolve().parent


OPS = {
    "!=": "__ne__",
    "==": "__eq__",
    ">": "__gt__",
    ">=": "__ge__",
    "<": "__lt__",
    "<=": "__le__",
}


class SignalOnline:
    def __init__(
        self,
        name="",
        raw=None,
        scaled=None,
        unit="",
        entry=(),
        conversion=None,
        exists=True,
        format="phys",
        color="#505050",
        y_range=(0, 100),
    ):
        self.name = name
        self.raw = raw
        self.scaled = scaled
        self.unit = unit
        self.entry = entry
        self.conversion = conversion
        self.exists = exists
        self.configured_from_device = True
        self.format = format
        self.y_range = y_range

        color = color or "#505050"
        self.color = fn.mkColor(color)

    @property
    def origin_mdf(self):
        return ""

    @origin_mdf.setter
    def origin_mdf(self, value):
        pass

    @property
    def origin_uuid(self):
        return self.entry[0]

    @origin_uuid.setter
    def origin_uuid(self, value):
        self.entry = (value, self.name)

    def reset(self):
        self.raw = None
        self.scaled = None
        self.exists = True

    def update_values(self, values):
        self.raw = values[-1]
        if self.conversion:
            self.scaled = self.conversion.convert(values[-1:], as_bytes=True)[0]
        else:
            self.scaled = self.raw

    def __lt__(self, other):
        return self.name < other.name

    def get_value(self, index):
        if index == 0:
            return self.name
        elif index == 1:
            return self.raw
        elif index == 2:
            return self.scaled
        elif index == 3:
            return self.unit
        elif index == 4:
            return self.origin_uuid


class SignalOffline:
    def __init__(
        self,
        signal=None,
        exists=True,
    ):
        self.signal = signal
        self.exists = exists
        self.raw = None
        self.scaled = None
        self.last_timestamp = None
        self.entry = signal.entry
        self.name = signal.name
        self.unit = signal.unit
        self.format = getattr(signal, "format", "phys")

        if not hasattr(signal, "color"):
            self.color = fn.mkColor("#505050")

        if not hasattr(signal, "y_range"):
            self.y_range = (0, 100)

    @property
    def y_range(self):
        return self.signal.y_range

    @y_range.setter
    def y_range(self, value):
        self.signal.y_range = value

    @property
    def color(self):
        return self.signal.color

    @color.setter
    def color(self, value):
        self.signal.color = value

    @property
    def origin_mdf(self):
        return self.signal.origin_mdf

    @origin_mdf.setter
    def origin_mdf(self, value):
        self.signal.origin_mdf = value

    @property
    def origin_uuid(self):
        return self.signal.origin_uuid

    @origin_uuid.setter
    def origin_uuid(self, value):
        self.signal.origin_uuid = value

    def reset(self, exists=True):
        self.signal.samples = self.signal.samples[:0]
        self.signal.timestamps = self.signal.timestamps[:0]
        self.exists = exists
        self.raw = None
        self.scaled = None
        self.last_timestamp = None

    def __lt__(self, other):
        return self.name < other.name

    def set_timestamp(self, timestamp):
        if timestamp is not None and (self.last_timestamp is None or self.last_timestamp != timestamp):
            self.last_timestamp = timestamp

            sig = self.signal
            if sig.samples.size:
                idx = searchsorted(sig.timestamps, timestamp, side="right")
                idx -= 1
                idx = max(idx, 0)

                self.raw = sig.raw_samples[idx]
                self.scaled = sig.phys_samples[idx]

    def get_value(self, index, timestamp=None):
        self.set_timestamp(timestamp)
        if self.signal is not None:
            if index == 0:
                return self.signal.name
            elif index == 1:
                return self.raw
            elif index == 2:
                return self.scaled
            elif index == 3:
                return self.unit
            elif index == 4:
                return self.origin_uuid


class OnlineBackEnd:
    def __init__(self, signals, numeric):
        super().__init__()

        self.signals = signals or []
        self.map = None
        self.numeric = numeric

        self.sorted_column_index = 0
        self.sorting_enabled = True
        self.sort_reversed = False
        self.numeric_viewer = None

        self.update()

    def update_signal_origin_uuid(self, signal, origin_uuid):
        old_entry = signal.entry
        signal.origin_uuid = origin_uuid
        self.map[signal.entry] = signal
        del self.map[old_entry]

        self.numeric_viewer.dataView.ranges[signal.entry] = self.numeric_viewer.dataView.ranges[old_entry]
        del self.numeric_viewer.dataView.ranges[old_entry]

    def update(self, others=()):
        self.map = {signal.entry: signal for signal in self.signals}
        for signal in others:
            if signal.entry not in self.map:
                self.map[signal.entry] = signal
                self.signals.append(signal)

        self.sort()

    def sort_column(self, ix):
        if ix != self.sorted_column_index:
            self.sorted_column_index = ix
            self.sort_reversed = False
        else:
            self.sort_reversed = not self.sort_reversed

        self.sort()

    def color_same_origin_signals(self, origin_uuid="", color=""):
        pass

    def data_changed(self):
        self.refresh_ui()

    def move_rows(self, rows, target_row):
        if target_row == -1:
            sigs = [self.signals.pop(row) for row in rows]
            self.signals.extend(sigs)
        else:
            sig = self.signals[target_row]
            sigs = [self.signals.pop(row) for row in rows]

            idx = self.signals.index(sig)
            for sig in sigs:
                self.signals.insert(idx, sig)

        self.data_changed()

    def refresh_ui(self):
        if self.numeric is not None and self.numeric.mode == "offline":
            numeric = self.numeric
            numeric._min = float("inf")
            numeric._max = -float("inf")

            for sig in self.signals:
                if sig.samples.size:
                    numeric._min = min(self._min, sig.timestamps[0])
                    numeric._max = max(self._max, sig.timestamps[-1])

            if numeric._min == float("inf"):
                numeric._min = numeric._max = 0

            numeric._timestamp = numeric._min

            numeric.timestamp.setRange(numeric._min, numeric._max)
            numeric.min_t.setText(f"{numeric._min:.9f}s")
            numeric.max_t.setText(f"{numeric._max:.9f}s")
            numeric.set_timestamp(numeric._min)

        if self.numeric_viewer is not None:
            self.numeric_viewer.refresh_ui()

    def reorder(self, names):
        try:
            sigs = {sig.name: idx for idx, sig in enumerate(self.signals)}

            if len(sigs) == len(names):
                self.signals = [self.signals[sigs[name]] for name in names]

            self.data_changed()

        except:
            pass

    def sort(self):
        if not self.sorting_enabled:
            self.data_changed()
            return

        sorted_column_index = self.sorted_column_index

        if sorted_column_index == 0:
            self.signals = natsorted(self.signals, key=lambda x: x.name, reverse=self.sort_reversed)

        elif sorted_column_index in (1, 2):
            numeric = []
            string = []
            nones = []

            for signal in self.signals:
                value = signal.get_value(sorted_column_index)
                if value is None:
                    nones.append(signal)
                elif isinstance(value, (np.flexible, bytes)):
                    string.append(signal)
                else:
                    numeric.append(signal)

            self.signals = [
                *sorted(
                    numeric,
                    key=lambda x: x.get_value(sorted_column_index),
                    reverse=self.sort_reversed,
                ),
                *sorted(
                    string,
                    key=lambda x: x.get_value(sorted_column_index),
                    reverse=self.sort_reversed,
                ),
                *natsorted(nones, key=lambda x: x.name, reverse=self.sort_reversed),
            ]

        elif sorted_column_index == 3:
            self.signals = natsorted(self.signals, key=lambda x: x.unit, reverse=self.sort_reversed)

        self.data_changed()

    def set_values(self, values=None):
        map_ = self.map
        if values:
            for entry, vals in values.items():
                sig = map_[entry]
                sig.update_values(vals)

            if self.sorted_column_index in (1, 2):
                self.sort()
            else:
                self.data_changed()

    def shift_same_origin_signals(self, origin_uuid="", delta=0.0, absolute=False):
        pass

    def update_missing_signals(self, uuids=()):
        pass

    def reset(self):
        for sig in self.signals:
            sig.reset()
        self.data_changed()

    def __len__(self):
        return len(self.signals)

    def does_not_exist(self, entry, exists):
        self.map[entry].exists = exists

    def get_signal_value(self, signal, column):
        return signal.get_value(column)

    def set_format(self, fmt, rows):
        for row in rows:
            self.signals[row].format = fmt


class OfflineBackEnd:
    def __init__(self, signals, numeric):
        super().__init__()

        self.timestamp = None

        self.signals = signals or []
        self.map = None
        self.numeric = numeric

        self.sorted_column_index = 0
        self.sorting_enabled = True
        self.sort_reversed = False
        self.numeric_viewer = None

        self.timebase = np.array([])

    def update(self, others=()):
        self.map = {signal.entry: signal for signal in self.signals}
        for signal in others:
            if signal.entry not in self.map:
                self.map[signal.entry] = signal
                self.signals.append(signal)

        if self.signals:
            timestamps = {id(signal.signal.timestamps): signal.signal.timestamps for signal in self.signals}
            timestamps = list(timestamps.values())
            self.timebase = np.unique(np.concatenate(timestamps))
        else:
            self.timebase = np.array([])

        self.sort()

    def sort_column(self, ix):
        if ix != self.sorted_column_index:
            self.sorted_column_index = ix
            self.sort_reversed = False
        else:
            self.sort_reversed = not self.sort_reversed

        self.sort()

    def data_changed(self):
        self.refresh_ui()

    def move_rows(self, rows, target_row):
        if target_row == -1:
            sigs = [self.signals.pop(row) for row in rows]
            self.signals.extend(sigs)
        else:
            sig = self.signals[target_row]
            sigs = [self.signals.pop(row) for row in rows]

            idx = self.signals.index(sig)
            for sig in sigs:
                self.signals.insert(idx, sig)

        self.data_changed()

    def refresh_ui(self):
        if self.numeric_viewer is not None:
            self.numeric_viewer.refresh_ui()

    def reorder(self, names):
        try:
            sigs = {sig.name: idx for idx, sig in enumerate(self.signals)}

            if len(sigs) == len(names):
                self.signals = [self.signals[sigs[name]] for name in names]

            self.data_changed()

        except:
            pass

    def sort(self):
        if not self.sorting_enabled:
            self.data_changed()
            return

        sorted_column_index = self.sorted_column_index

        if sorted_column_index == 0:
            self.signals = natsorted(self.signals, key=lambda x: (x.name, x.origin_uuid), reverse=self.sort_reversed)

        elif sorted_column_index in (1, 2):
            numeric = []
            string = []
            nones = []

            for signal in self.signals:
                value = signal.get_value(sorted_column_index, self.timestamp)
                if value is None:
                    nones.append(signal)
                elif isinstance(value, (np.flexible, bytes)):
                    string.append(signal)
                else:
                    numeric.append(signal)

            self.signals = [
                *sorted(
                    numeric,
                    key=lambda x: x.get_value(sorted_column_index),
                    reverse=self.sort_reversed,
                ),
                *sorted(
                    string,
                    key=lambda x: x.get_value(sorted_column_index),
                    reverse=self.sort_reversed,
                ),
                *natsorted(nones, key=lambda x: x.name, reverse=self.sort_reversed),
            ]

        elif sorted_column_index == 3:
            self.signals = natsorted(self.signals, key=lambda x: x.unit, reverse=self.sort_reversed)

        self.data_changed()

    def color_same_origin_signals(self, origin_uuid="", color=""):
        for signal in self.signals:
            if signal.origin_uuid == origin_uuid:
                signal.color = color

        self.data_changed()

    def get_timestamp(self, stamp):
        max_idx = len(self.timebase) - 1
        if max_idx == -1:
            return stamp

        idx = np.searchsorted(self.timebase, stamp)
        idx = min(idx, max_idx)

        return self.timebase[idx]

    def set_timestamp(self, stamp):
        self.timestamp = stamp
        if self.sorted_column_index in (1, 2):
            self.sort()
        else:
            self.data_changed()

    def shift_same_origin_signals(self, origin_uuid="", delta=0.0, absolute=False):
        for signal in self.signals:
            if signal.origin_uuid == origin_uuid:
                if not absolute:
                    signal.signal.timestamps = signal.signal.timestamps + delta
                else:
                    if len(signal.signal.timestamps):
                        signal.signal.timestamps = signal.signal.timestamps - signal.signal.timestamps[0] + delta

        self.data_changed()

    def update_missing_signals(self, uuids=()):
        for signal in self.signals:
            if signal.origin_uuid not in uuids:
                signal.reset(exists=False)

    def reset(self):
        for sig in self.signals:
            sig.reset()
        self.data_changed()

    def __len__(self):
        return len(self.signals)

    def does_not_exist(self, entry, exists):
        self.map[entry].exists = exists

    def get_signal_value(self, signal, column):
        return signal.get_value(column, self.timestamp)

    def set_format(self, fmt, rows):
        for row in rows:
            self.signals[row].format = fmt


class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, parent, background_color, font_color):
        super().__init__(parent)
        self.numeric_viewer = parent
        self.backend = parent.backend
        self.view = None
        self.format = "Physical"
        self.float_precision = -1
        self.background_color = background_color
        self.font_color = font_color

    def headerData(self, section, orientation, role=None):
        pass

    def columnCount(self, parent=None):
        return 5

    def rowCount(self, parent=None):
        return len(self.backend)

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        row = index.row()
        col = index.column()

        signal = self.backend.signals[row]
        cell = self.backend.get_signal_value(signal, col)

        match role:
            case QtCore.Qt.ItemDataRole.DisplayRole:
                if cell is None:
                    return "●"
                elif isinstance(cell, (bytes, np.bytes_)):
                    return cell.decode("utf-8", "replace")
                elif isinstance(cell, str):
                    return cell
                elif isinstance(cell, (np.ndarray, np.record, np.recarray)):
                    return str(cell[0])
                else:
                    if np.isnan(cell):
                        return "NaN"
                    else:
                        return value_as_str(cell, signal.format, None, self.float_precision)

            case QtCore.Qt.ItemDataRole.BackgroundRole:
                channel_ranges = self.view.ranges[signal.entry]
                raw_cell = self.backend.get_signal_value(signal, 1)
                scaled_cell = self.backend.get_signal_value(signal, 2)

                try:
                    scaled_value = float(scaled_cell)
                    value = scaled_value
                except:
                    scaled_value = str(scaled_cell)

                    try:
                        raw_value = float(raw_cell)
                        value = raw_value
                    except:
                        value = scaled_value

                new_background_color, new_font_color = get_colors_using_ranges(
                    value,
                    ranges=channel_ranges,
                    default_background_color=self.background_color,
                    default_font_color=signal.color,
                )

                return new_background_color if new_background_color != self.background_color else None

            case QtCore.Qt.ItemDataRole.ForegroundRole:
                channel_ranges = self.view.ranges[signal.entry]
                raw_cell = self.backend.get_signal_value(signal, 1)
                scaled_cell = self.backend.get_signal_value(signal, 2)

                try:
                    scaled_value = float(scaled_cell)
                    value = scaled_value
                except:
                    scaled_value = str(scaled_cell)

                    try:
                        raw_value = float(raw_cell)
                        value = raw_value
                    except:
                        value = scaled_value

                new_background_color, new_font_color = get_colors_using_ranges(
                    value,
                    ranges=channel_ranges,
                    default_background_color=self.background_color,
                    default_font_color=signal.color,
                )

                return new_font_color

            case QtCore.Qt.ItemDataRole.TextAlignmentRole:
                if col:
                    return int(QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter)
                else:
                    return int(QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignVCenter)

            case QtCore.Qt.ItemDataRole.DecorationRole:
                if col == 0:
                    if not signal.exists:
                        icon = utils.ERROR_ICON
                        if icon is None:
                            utils.ERROR_ICON = QtGui.QIcon()
                            utils.ERROR_ICON.addPixmap(
                                QtGui.QPixmap(":/error.png"),
                                QtGui.QIcon.Mode.Normal,
                                QtGui.QIcon.State.Off,
                            )

                            utils.NO_ERROR_ICON = QtGui.QIcon()

                            icon = utils.ERROR_ICON
                    else:
                        icon = utils.NO_ERROR_ICON
                        if icon is None:
                            utils.ERROR_ICON = QtGui.QIcon()
                            utils.ERROR_ICON.addPixmap(
                                QtGui.QPixmap(":/error.png"),
                                QtGui.QIcon.Mode.Normal,
                                QtGui.QIcon.State.Off,
                            )

                            utils.NO_ERROR_ICON = QtGui.QIcon()

                            icon = utils.NO_ERROR_ICON

                    return icon

                elif col in (1, 2):
                    has_ranges = bool(self.view.ranges.get(signal.entry, False))
                    if has_ranges:
                        icon = utils.RANGE_INDICATOR_ICON
                        if icon is None:
                            utils.RANGE_INDICATOR_ICON = QtGui.QIcon()
                            utils.RANGE_INDICATOR_ICON.addPixmap(
                                QtGui.QPixmap(":/paint.png"),
                                QtGui.QIcon.Mode.Normal,
                                QtGui.QIcon.State.Off,
                            )

                            utils.NO_ERROR_ICON = QtGui.QIcon()
                            utils.NO_ICON = QtGui.QIcon()

                            icon = utils.RANGE_INDICATOR_ICON
                    else:
                        icon = utils.NO_ERROR_ICON
                        if icon is None:
                            utils.RANGE_INDICATOR_ICON = QtGui.QIcon()
                            utils.RANGE_INDICATOR_ICON.addPixmap(
                                QtGui.QPixmap(":/paint.png"),
                                QtGui.QIcon.Mode.Normal,
                                QtGui.QIcon.State.Off,
                            )

                            utils.NO_ERROR_ICON = QtGui.QIcon()
                            utils.NO_ICON = QtGui.QIcon()

                            icon = utils.NO_ERROR_ICON

                    return icon

            case QtCore.Qt.ItemDataRole.ToolTipRole:
                if signal:
                    return f"Origin = {signal.origin_uuid or 'unknown'}\nMDF = {signal.origin_mdf or 'unknown'}"

    def flags(self, index):
        return (
            QtCore.Qt.ItemFlag.ItemIsEnabled
            | QtCore.Qt.ItemFlag.ItemIsSelectable
            | QtCore.Qt.ItemFlag.ItemIsDragEnabled
            | QtCore.Qt.ItemFlag.ItemIsDropEnabled
        )

    def dropMimeData(self, data, action, row, column, parent):
        def moved_rows(data):
            rows = set()
            ds = QtCore.QDataStream(data.data("application/x-qabstractitemmodeldatalist"))
            while not ds.atEnd():
                row = ds.readInt32()
                ds.readInt32()
                map_items = ds.readInt32()
                for i in range(map_items):
                    ds.readInt32()
                    ds.readQVariant()

                rows.add(row)

            return sorted(rows, reverse=True)

        self.backend.move_rows(moved_rows(data), parent.row())

    def supportedDropActions(self) -> bool:
        return QtCore.Qt.DropAction.MoveAction | QtCore.Qt.DropAction.CopyAction

    def set_format(self, fmt, indexes):
        if fmt not in ("phys", "hex", "bin", "ascii"):
            return

        self.format = fmt

        rows = {index.row() for index in indexes}

        self.backend.set_format(fmt, rows)


class TableView(QtWidgets.QTableView):
    add_channels_request = QtCore.Signal(list)

    def __init__(self, parent):
        super().__init__(parent)
        self.numeric_viewer = parent
        self.backend = parent.backend

        self.ranges = {}

        self._backgrund_color = self.palette().color(QtGui.QPalette.ColorRole.Window)
        self._font_color = self.palette().color(QtGui.QPalette.ColorRole.WindowText)

        model = TableModel(parent, self._backgrund_color, self._font_color)
        self.setModel(model)
        model.view = self

        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        self.doubleClicked.connect(self.edit_ranges)

        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.InternalMove)

        self.double_clicked_enabled = True

    def sizeHint(self):
        width = 2 * self.frameWidth()
        for i in range(self.model().columnCount()):
            width += self.columnWidth(i)

        height = 2 * self.frameWidth()
        height += 24 * self.model().rowCount()

        return QtCore.QSize(width, height)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if key == QtCore.Qt.Key.Key_Delete and modifiers == QtCore.Qt.KeyboardModifier.NoModifier:
            event.accept()
            selected_items = {index.row() for index in self.selectedIndexes() if index.isValid()}

            for row in sorted(selected_items, reverse=True):
                signal = self.backend.signals.pop(row)
                del self.backend.map[signal.entry]

            self.backend.update()

        elif key == QtCore.Qt.Key.Key_R and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            event.accept()

            selected_items = {index.row() for index in self.selectedIndexes() if index.isValid()}

            if selected_items:
                ranges = []

                for row in selected_items:
                    signal = self.backend.signals[row]
                    if self.ranges[signal.entry]:
                        ranges.extend(self.ranges[signal.entry])

                dlg = RangeEditor(
                    "<selected items>",
                    "",
                    ranges=unique_ranges(ranges),
                    parent=self,
                    brush=True,
                )
                dlg.exec_()
                if dlg.pressed_button == "apply":
                    ranges = dlg.result
                    for row in selected_items:
                        signal = self.backend.signals[row]
                        self.ranges[signal.entry] = copy_ranges(ranges)

                    self.backend.update()

        elif (
            modifiers == (QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier)
            and key == QtCore.Qt.Key.Key_C
        ):
            event.accept()

            selected_items = {index.row() for index in self.selectedIndexes() if index.isValid()}

            if not selected_items:
                return
            else:
                precision = self.model().float_precision
                row = list(selected_items)[0]
                signal = self.backend.signals[row]

                info = {
                    "format": signal.format,
                    "ranges": self.ranges[signal.entry],
                    "type": "channel",
                    "color": signal.color,
                    "precision": precision,
                    "ylink": False,
                    "individual_axis": False,
                    "y_range": signal.y_range,
                    "origin_uuid": signal.origin_uuid,
                }

                QtWidgets.QApplication.instance().clipboard().setText(json.dumps(info, cls=ExtendedJsonEncoder))

        elif (
            modifiers == (QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier)
            and key == QtCore.Qt.Key.Key_V
        ):
            event.accept()

            info = QtWidgets.QApplication.instance().clipboard().text()
            selected_items = {index.row() for index in self.selectedIndexes() if index.isValid()}

            if not selected_items:
                return

            try:
                info = json.loads(info, cls=ExtendedJsonDecoder)
            except:
                print(format_exc())
            else:
                for row in selected_items:
                    signal = self.backend.signals[row]

                    signal.format = info["format"]
                    signal.color = fn.mkColor(info["color"])
                    signal.y_range = info["y_range"]
                    signal.origin_uuid = info["origin_uuid"]
                    self.ranges[signal.entry] = copy_ranges(info["ranges"])

                self.backend.update()

        elif modifiers == QtCore.Qt.KeyboardModifier.NoModifier and key == QtCore.Qt.Key.Key_C:
            event.accept()

            selected_items = [index.row() for index in self.selectedIndexes() if index.isValid()]

            if selected_items:
                signal = self.backend.signals[selected_items[0]]
                color = signal.color

                color = QtWidgets.QColorDialog.getColor(color, parent=self)
                if color.isValid():
                    for row in set(selected_items):
                        signal = self.backend.signals[row]
                        signal.color = color

        elif modifiers == QtCore.Qt.KeyboardModifier.ControlModifier and key == QtCore.Qt.Key.Key_N:
            event.accept()
            selected_items = []

            for index in self.selectedIndexes():
                if not index.isValid():
                    continue

                if (row := index.row()) not in selected_items:
                    selected_items.append(row)

            if not selected_items:
                return
            else:
                text = "\n".join(self.backend.signals[row].name for row in selected_items)

            QtWidgets.QApplication.instance().clipboard().setText(text)

        elif modifiers == QtCore.Qt.KeyboardModifier.ControlModifier and key == QtCore.Qt.Key.Key_C:
            event.accept()

            selected_items = []

            for index in self.selectedIndexes():
                if not index.isValid():
                    continue

                if (row := index.row()) not in selected_items:
                    selected_items.append(row)

            data = []
            numeric_mode = self.backend.numeric.mode

            for row in selected_items:
                signal = self.backend.signals[row]

                entry = signal.entry if numeric_mode == "online" else signal.signal.entry

                group_index, channel_index = entry

                info = {
                    "name": signal.name,
                    "computation": {},
                    "computed": False,
                    "group_index": group_index,
                    "channel_index": channel_index,
                    "ranges": self.ranges[signal.entry],
                    "origin_uuid": str(entry[0]) if numeric_mode == "online" else signal.signal.origin_uuid,
                    "type": "channel",
                    "uuid": os.urandom(6).hex(),
                    "color": signal.color,
                }

                data.append(info)

            data = substitude_mime_uuids(data, None, force=True)
            QtWidgets.QApplication.instance().clipboard().setText(json.dumps(data, cls=ExtendedJsonEncoder))

        elif modifiers == QtCore.Qt.KeyboardModifier.ControlModifier and key == QtCore.Qt.Key.Key_V:
            event.accept()
            try:
                data = QtWidgets.QApplication.instance().clipboard().text()
                data = json.loads(data, cls=ExtendedJsonDecoder)
                data = substitude_mime_uuids(data, random_uuid=True)
                self.add_channels_request.emit(data)
            except:
                pass

        else:
            super().keyPressEvent(event)

    def startDrag(self, supportedActions):
        indexes = self.selectedIndexes()
        if not self.backend.sorting_enabled:
            mime_data = self.model().mimeData(indexes)
        else:
            mime_data = QtCore.QMimeData()

        selected_items = []

        for index in self.selectedIndexes():
            if not index.isValid():
                continue

            if (row := index.row()) not in selected_items:
                selected_items.append(row)

        data = []
        numeric_mode = self.backend.numeric.mode

        for row in selected_items:
            signal = self.backend.signals[row]

            entry = signal.entry if numeric_mode == "online" else signal.signal.entry

            *_, group_index, channel_index = entry

            info = {
                "name": signal.name,
                "computation": {},
                "computed": False,
                "group_index": group_index,
                "channel_index": channel_index,
                "ranges": self.ranges[signal.entry],
                "origin_uuid": str(entry[0]),
                "type": "channel",
                "uuid": os.urandom(6).hex(),
                "color": signal.color,
            }

            data.append(info)

        data = json.dumps(data, cls=ExtendedJsonEncoder).encode("utf-8")

        mime_data.setData("application/octet-stream-asammdf", QtCore.QByteArray(data))

        drag = QtGui.QDrag(self)
        drag.setMimeData(mime_data)
        drag.exec(supportedActions)

    def dragEnterEvent(self, e):
        e.accept()

    def dropEvent(self, e):
        if e.source() is self:
            if e.mimeData().hasFormat("application/x-qabstractitemmodeldatalist"):
                e.mimeData().removeFormat("application/octet-stream-asammdf")
                super().dropEvent(e)
            else:
                e.ignore()
            self.clearSelection()
        else:
            data = e.mimeData()
            if data.hasFormat("application/octet-stream-asammdf"):
                names = extract_mime_names(data)
                self.add_channels_request.emit(names)
                e.accept()
            else:
                e.ignore()

    def edit_ranges(self, index):
        if not self.double_clicked_enabled or not index.isValid():
            return

        row = index.row()
        signal = self.backend.signals[row]

        dlg = RangeEditor(signal.name, signal.unit, self.ranges[signal.entry], parent=self, brush=True)
        dlg.exec_()
        if dlg.pressed_button == "apply":
            ranges = dlg.result
            self.ranges[signal.entry] = ranges

    def set_format(self, fmt):
        indexes = self.selectedIndexes()
        self.model().set_format(fmt, indexes)


class HeaderModel(QtCore.QAbstractTableModel):
    def __init__(self, parent):
        super().__init__(parent)
        self.backend = parent.backend

    def columnCount(self, parent=None):
        return 5

    def rowCount(self, parent=None):
        return 1  # 1?

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        col = index.column()

        names = ["Name", "Raw", "Scaled", "Unit", "Origin"]

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            return names[col]

        elif role == QtCore.Qt.ItemDataRole.DecorationRole:
            if not self.backend.sorting_enabled or col != self.backend.sorted_column_index:
                return
            else:
                if self.backend.sort_reversed:
                    icon = QtGui.QIcon(":/sort-descending.png")
                else:
                    icon = QtGui.QIcon(":/sort-ascending.png")

                return icon

        elif role == QtCore.Qt.ItemDataRole.TextAlignmentRole:
            return QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignVCenter

    def headerData(self, section, orientation, role=None):
        pass


class HeaderView(QtWidgets.QTableView):
    sorting_changed = QtCore.Signal(int)
    NameColumn = 0
    RawColumn = 1
    ScaledColumn = 2
    UnitColumn = 3
    OriginColumn = 4

    def __init__(self, parent):
        super().__init__(parent)
        self.numeric_viewer = parent
        self.backend = parent.backend

        self.table = parent.dataView
        self.setModel(HeaderModel(parent))
        self.padding = 10

        self.header_cell_being_resized = None
        self.header_being_resized = False

        self.setMouseTracking(True)
        self.viewport().setMouseTracking(True)
        self.viewport().installEventFilter(self)

        self.setIconSize(QtCore.QSize(16, 16))
        self.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)
        )
        self.setWordWrap(False)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        font = QtGui.QFont()
        font.setBold(True)
        self.setFont(font)

        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.resize(self.sizeHint())

        self.columns_width = {
            self.RawColumn: self.columnWidth(self.RawColumn),
            self.ScaledColumn: self.columnWidth(self.ScaledColumn),
            self.UnitColumn: self.columnWidth(self.UnitColumn),
            self.OriginColumn: self.columnWidth(self.OriginColumn),
        }

    def all_columns_width(self):
        widths = []
        for column in (self.NameColumn, self.RawColumn, self.ScaledColumn, self.UnitColumn, self.OriginColumn):
            if self.isColumnHidden(column):
                widths.append(self.columns_width.get(column, 100))
            else:
                widths.append(self.columnWidth(column))
        return widths

    def columns_visibility(self):
        return {
            "raw": not self.isColumnHidden(self.RawColumn),
            "scaled": not self.isColumnHidden(self.ScaledColumn),
            "unit": not self.isColumnHidden(self.UnitColumn),
            "origin": not self.isColumnHidden(self.OriginColumn),
        }

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        super().showEvent(a0)
        self.initial_size = self.size()

    def sorting(self):
        return {
            "sort_column": self.backend.sorted_column_index,
            "enabled": self.backend.sorting_enabled,
            "reversed": self.backend.sort_reversed,
        }

    def mouseDoubleClickEvent(self, event):
        point = event.pos()
        ix = self.indexAt(point)
        col = ix.column()
        if event.button() == QtCore.Qt.MouseButton.LeftButton and self.backend.sorting_enabled:
            self.backend.sort_column(col)
            self.sorting_changed.emit(col)
        else:
            super().mouseDoubleClickEvent(event)

    def eventFilter(self, object: QtCore.QObject, event: QtCore.QEvent):
        if event.type() in [
            QtCore.QEvent.Type.MouseButtonPress,
            QtCore.QEvent.Type.MouseButtonRelease,
            QtCore.QEvent.Type.MouseButtonDblClick,
            QtCore.QEvent.Type.MouseMove,
        ]:
            return self.manage_resizing(object, event)

        return False

    def manage_resizing(self, object: QtCore.QObject, event: QtCore.QEvent):
        def over_header_cell_edge(mouse_position, margin=3):
            x = mouse_position
            if self.columnAt(x - margin) != self.columnAt(x + margin):
                if self.columnAt(x + margin) == 0:
                    return None
                else:
                    return self.columnAt(x - margin)
            else:
                return None

        mouse_position = event.pos().x()
        orthogonal_mouse_position = event.pos().y()

        if over_header_cell_edge(mouse_position) is not None:
            self.viewport().setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.SplitHCursor))

        else:
            self.viewport().setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.ArrowCursor))

        if event.type() == QtCore.QEvent.Type.MouseButtonPress:
            if over_header_cell_edge(mouse_position) is not None:
                self.header_cell_being_resized = over_header_cell_edge(mouse_position)
                return True
            else:
                self.header_cell_being_resized = None

        if event.type() == QtCore.QEvent.Type.MouseButtonRelease:
            self.header_cell_being_resized = None
            self.header_being_resized = False

        if event.type() == QtCore.QEvent.Type.MouseButtonDblClick:
            if over_header_cell_edge(mouse_position) is not None:
                header_index = over_header_cell_edge(mouse_position)
                self.numeric_viewer.auto_size_column(header_index)
                return True

        if event.type() == QtCore.QEvent.Type.MouseMove:
            if self.header_cell_being_resized is not None:
                size = mouse_position - self.columnViewportPosition(self.header_cell_being_resized)
                if size > 10:
                    self.setColumnWidth(self.header_cell_being_resized, size)
                    self.numeric_viewer.dataView.setColumnWidth(self.header_cell_being_resized, size)

                    self.updateGeometry()
                    self.numeric_viewer.dataView.updateGeometry()
                return True

            elif self.header_being_resized:
                size = orthogonal_mouse_position - self.geometry().top()
                self.setFixedHeight(max(size, self.initial_size.height()))

                self.updateGeometry()
                self.numeric_viewer.dataView.updateGeometry()
                return True

        return False

    def sizeHint(self):
        width = self.table.sizeHint().width() + self.verticalHeader().width()
        height = 16 + self.font().pointSize() + 2 * self.frameWidth()

        return QtCore.QSize(width, height)

    def toggle_column(self, checked, column):
        if not checked:
            self.columns_width[column] = self.columnWidth(column)

        self.setColumnHidden(column, not checked)
        self.numeric_viewer.dataView.setColumnHidden(column, not checked)

        if checked:
            self.setColumnWidth(column, self.columns_width[column])
            self.numeric_viewer.dataView.setColumnWidth(column, self.columns_width[column])

        self.updateGeometry()
        self.numeric_viewer.dataView.updateGeometry()

    def toggle_sorting(self, checked):
        self.backend.sorting_enabled = checked
        self.backend.sort()

    def minimumSizeHint(self):
        return QtCore.QSize(50, self.sizeHint().height())


class NumericViewer(QtWidgets.QWidget):
    def __init__(self, backend):
        super().__init__()

        backend.numeric_viewer = self
        self.backend = backend

        self.dataView = TableView(parent=self)

        self.columnHeader = HeaderView(parent=self)

        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.setLayout(self.gridLayout)

        self.dataView.horizontalScrollBar().valueChanged.connect(self.columnHeader.horizontalScrollBar().setValue)

        self.columnHeader.horizontalScrollBar().valueChanged.connect(self.dataView.horizontalScrollBar().setValue)

        # self.dataView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        # self.dataView.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.gridLayout.addWidget(self.columnHeader, 0, 0)
        self.gridLayout.addWidget(self.dataView, 1, 0)
        # self.gridLayout.addWidget(self.dataView.horizontalScrollBar(), 2, 0, 1, 1)
        # self.gridLayout.addWidget(self.dataView.verticalScrollBar(), 1, 1, 1, 1)

        self.dataView.verticalScrollBar().setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Ignored)
        )
        self.dataView.horizontalScrollBar().setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Fixed)
        )

        self.gridLayout.setColumnStretch(0, 1)
        self.gridLayout.setRowStretch(1, 1)

        self.columnHeader.setSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)

        self.default_row_height = 24
        self.set_styles()

        for column_index in range(self.columnHeader.model().columnCount()):
            self.auto_size_column(column_index)

        self.columnHeader.horizontalHeader().setStretchLastSection(True)

        self.columnHeader.horizontalHeader().sectionResized.connect(self.update_horizontal_scroll)

        self.columnHeader.horizontalHeader().setMinimumSectionSize(1)
        self.dataView.horizontalHeader().setMinimumSectionSize(1)

        self.show()

    def set_styles(self):
        self.dataView.verticalHeader().setDefaultSectionSize(self.default_row_height)
        self.dataView.verticalHeader().setMinimumSectionSize(self.default_row_height)
        self.dataView.verticalHeader().setMaximumSectionSize(self.default_row_height)
        self.dataView.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Fixed)
        self.columnHeader.verticalHeader().setDefaultSectionSize(self.default_row_height)
        self.columnHeader.verticalHeader().setMinimumSectionSize(self.default_row_height)
        self.columnHeader.verticalHeader().setMaximumSectionSize(self.default_row_height)
        self.columnHeader.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Fixed)

    def auto_size_header(self):
        s = 0
        for i in range(self.columnHeader.model().columnCount()):
            s += self.auto_size_column(i)

        delta = int((self.dataView.viewport().size().width() - s) // 4)

        if delta > 0:
            for i in range(self.columnHeader.model().columnCount()):
                self.auto_size_column(i, extra_padding=delta)
            # self.dataView.horizontalScrollBar().hide()
        else:
            self.dataView.horizontalScrollBar().show()

    def update_horizontal_scroll(self, *args):
        return

        s = 0
        for i in range(self.columnHeader.model().columnCount()):
            s += self.dataView.columnWidth(i) + self.dataView.frameWidth()

        if self.dataView.viewport().size().width() < s:
            self.dataView.horizontalScrollBar().show()
        else:
            self.dataView.horizontalScrollBar().hide()

    def auto_size_column(self, column_index, extra_padding=0):
        width = 0

        N = 100
        for i in range(self.dataView.model().rowCount())[:N]:
            mi = self.dataView.model().index(i, column_index)
            text = self.dataView.model().data(mi)
            w = self.dataView.fontMetrics().boundingRect(text.replace("\0", " ")).width()
            width = max(width, w)

        for i in range(self.columnHeader.model().rowCount()):
            mi = self.columnHeader.model().index(i, column_index)
            text = self.columnHeader.model().data(mi)
            w = self.columnHeader.fontMetrics().boundingRect(text.replace("\0", " ")).width()
            width = max(width, w)

        padding = 20
        width += padding + extra_padding

        self.columnHeader.setColumnWidth(column_index, width)
        self.dataView.setColumnWidth(column_index, self.columnHeader.columnWidth(column_index))

        self.dataView.updateGeometry()
        self.columnHeader.updateGeometry()

        return width

    def scroll_to_column(self, column=0):
        index = self.dataView.model().index(0, column)
        self.dataView.scrollTo(index)
        self.columnHeader.selectColumn(column)
        self.columnHeader.on_selectionChanged(force=True)

    def refresh_ui(self):
        self.models = []
        self.models += [
            self.dataView.model(),
            self.columnHeader.model(),
        ]

        for model in self.models:
            model.beginResetModel()
            model.endResetModel()

        for view in [self.columnHeader, self.dataView]:
            view.updateGeometry()


class Numeric(Ui_NumericDisplay, QtWidgets.QWidget):
    add_channels_request = QtCore.Signal(list)
    timestamp_changed_signal = QtCore.Signal(object, float)

    def __init__(
        self,
        channels=None,
        format=None,
        mode="offline",
        float_precision=None,
        owner=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.mode = mode
        self.owner = owner

        self.lock = Lock()
        self.visible_entries_modified = True

        self._settings = QtCore.QSettings()

        if mode == "offline":
            backend = OfflineBackEnd(None, self)
        else:
            backend = OnlineBackEnd(None, self)
        self.channels = NumericViewer(backend)
        self.backend = backend

        self.channels.dataView.ranges = {}

        self.main_layout.insertWidget(0, self.channels)
        self.main_layout.setStretch(0, 1)

        self.float_precision.addItems(["Full"] + [f"{i} decimals" for i in range(16)])

        self.float_precision.currentIndexChanged.connect(self.set_float_precision)

        format = format or self._settings.value("numeric_format", "Physical")
        if format not in ("Physical", "Hex", "Binary", "Ascii"):
            format = "Physical"
            self._settings.setValue("numeric_format", format)

        if float_precision is None:
            float_precision = self._settings.value("numeric_float_precision", -1, type=int)
        self.float_precision.setCurrentIndex(float_precision + 1)

        self.timebase = np.array([])
        self._timestamp = None

        if channels:
            self.add_new_channels(channels)

        self.channels.dataView.add_channels_request.connect(self.add_channels_request)
        self.channels.dataView.verticalScrollBar().valueChanged.connect(self.reset_visible_entries)
        self.channels.columnHeader.sorting_changed.connect(self.reset_visible_entries)

        self.channels.auto_size_header()
        self.double_clicked_enabled = True

        self.pattern = {}

        if self.mode == "offline":
            self.timestamp.valueChanged.connect(self._timestamp_changed)
            self.timestamp_slider.valueChanged.connect(self._timestamp_slider_changed)

            self._inhibit = False

            self.forward.clicked.connect(self.search_forward)
            self.backward.clicked.connect(self.search_backward)
            self.op.addItems([">", ">=", "<", "<=", "==", "!="])

            self.time_group.setHidden(True)
            self.search_group.setHidden(True)

            self.toggle_controls_btn.clicked.connect(self.toggle_controls)
        else:
            self.toggle_controls_btn.setHidden(True)
            self.time_group.setHidden(True)
            self.search_group.setHidden(True)

        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_menu)

    def show_menu(self, position):
        count = len(self.channels.backend)

        header = self.channels.columnHeader

        menu = QtWidgets.QMenu()
        menu.addAction(f"{count} rows in the numeric window")
        menu.addSeparator()

        action = QtGui.QAction("Sorting", menu)
        action.setCheckable(True)
        action.setChecked(self.channels.backend.sorting_enabled)
        action.toggled.connect(header.toggle_sorting)
        menu.addAction(action)

        menu.addAction("Automatic set columns width")
        menu.addSeparator()

        action = QtGui.QAction("Raw Column", menu)
        action.setCheckable(True)
        action.setChecked(not header.isColumnHidden(header.RawColumn))
        action.toggled.connect(partial(header.toggle_column, column=header.RawColumn))
        menu.addAction(action)

        action = QtGui.QAction("Scaled Column", menu)
        action.setCheckable(True)
        action.setChecked(not header.isColumnHidden(header.ScaledColumn))
        action.toggled.connect(partial(header.toggle_column, column=header.ScaledColumn))
        menu.addAction(action)

        action = QtGui.QAction("Unit Column", menu)
        action.setCheckable(True)
        action.setChecked(not header.isColumnHidden(header.UnitColumn))
        action.toggled.connect(partial(header.toggle_column, column=header.UnitColumn))
        menu.addAction(action)

        action = QtGui.QAction("Origin Column", menu)
        action.setCheckable(True)
        action.setChecked(not header.isColumnHidden(header.OriginColumn))
        action.toggled.connect(partial(header.toggle_column, column=header.OriginColumn))
        menu.addAction(action)

        action = QtGui.QAction("Hide header and controls", menu)
        action.setCheckable(True)
        action.setChecked(header.isHidden())
        menu.addAction(action)

        menu.addSeparator()

        submenu = QtWidgets.QMenu("Copy names")
        submenu.setIcon(QtGui.QIcon(":/copy.png"))
        submenu.addAction("Copy names [Ctrl+N]")
        submenu.addAction("Copy names and values")
        menu.addMenu(submenu)

        submenu = QtWidgets.QMenu("Display structure")
        submenu.setIcon(QtGui.QIcon(":/structure.png"))
        submenu.addAction("Copy display properties [Ctrl+Shift+C]")
        submenu.addAction("Paste display properties [Ctrl+Shift+V]")
        submenu.addAction("Copy channel structure [Ctrl+C]")
        submenu.addAction("Paste channel structure [Ctrl+V]")
        menu.addMenu(submenu)

        menu.addSeparator()

        submenu = QtWidgets.QMenu("Edit")
        submenu.setIcon(QtGui.QIcon(":/edit.png"))

        submenu.addAction("Set color [C]")
        submenu.addAction("Set random color")
        submenu.addAction("Set color ranges [Ctrl+R]")
        menu.addMenu(submenu)

        menu.addSeparator()

        submenu = QtWidgets.QMenu("Display mode")
        submenu.addAction("Ascii\t[Ctrl+T]")
        submenu.addAction("Bin\t[Ctrl+B]")
        submenu.addAction("Hex\t[Ctrl+H]")
        submenu.addAction("Physical\t[Ctrl+P]")
        menu.addMenu(submenu)

        menu.addSeparator()
        menu.addAction(QtGui.QIcon(":/erase.png"), "Delete [Del]")

        action = menu.exec_(self.mapToGlobal(position))

        if action is None:
            return

        action_text = action.text()

        if action_text == "Copy names [Ctrl+N]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_N, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)

        elif action_text == "Copy names and values":
            texts = []
            precision = self.float_precision.currentIndex() - 1

            t = self.timestamp.value()

            if precision == -1:
                t = str(t)
            else:
                template = f"{{:.{precision}f}}"
                t = template.format(t)

            selected_items = []

            for index in self.channels.dataView.selectedIndexes():
                if not index.isValid():
                    continue

                if (row := index.row()) not in selected_items:
                    selected_items.append(row)

            model = self.channels.dataView.model()

            for row in selected_items:
                texts.append(
                    ", ".join(
                        [
                            model.data(model.createIndex(row, HeaderView.NameColumn)),
                            t,
                            f"{model.data(model.createIndex(row, HeaderView.RawColumn))}",
                            f"{model.data(model.createIndex(row, HeaderView.ScaledColumn))}{model.data(model.createIndex(row, HeaderView.UnitColumn))}",
                        ]
                    )
                )

            QtWidgets.QApplication.instance().clipboard().setText("\n".join(texts))

        elif action_text == "Copy channel structure [Ctrl+C]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_C,
                QtCore.Qt.KeyboardModifier.ControlModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Paste channel structure [Ctrl+V]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_V,
                QtCore.Qt.KeyboardModifier.ControlModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Copy display properties [Ctrl+Shift+C]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_C,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Paste display properties [Ctrl+Shift+V]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_V,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Copy display properties [Ctrl+Shift+C]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_C,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Paste display properties [Ctrl+Shift+V]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_V,
                QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Automatic set columns width":
            header.numeric_viewer.auto_size_header()
        elif action_text == "Hide header and controls":
            if action.isChecked():
                header.hide()
                self.controls.hide()
            else:
                header.show()
                self.controls.show()

        elif action_text == "Set color [C]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_C,
                QtCore.Qt.KeyboardModifier.NoModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Set color ranges [Ctrl+R]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                QtCore.Qt.Key.Key_R,
                QtCore.Qt.KeyboardModifier.ControlModifier,
            )
            self.keyPressEvent(event)

        elif action_text == "Set random color":
            selected_items = {index.row() for index in self.channels.dataView.selectedIndexes() if index.isValid()}

            for row in selected_items:
                while True:
                    rgb = os.urandom(3)
                    if 100 <= sum(rgb) <= 650:
                        break

                self.backend.signals[row].color = fn.mkColor(f"#{rgb.hex()}")

        elif action_text == "Ascii\t[Ctrl+T]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_T, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)
        elif action_text == "Bin\t[Ctrl+B]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_B, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)
        elif action_text == "Hex\t[Ctrl+H]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_H, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)
        elif action_text == "Physical\t[Ctrl+P]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_P, QtCore.Qt.KeyboardModifier.ControlModifier
            )
            self.keyPressEvent(event)

        elif action_text == "Delete [Del]":
            event = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress, QtCore.Qt.Key.Key_Delete, QtCore.Qt.KeyboardModifier.NoModifier
            )
            self.keyPressEvent(event)

    def add_new_channels(self, channels, mime_data=None):
        if self.mode == "online":
            others = []
            for index, sig in enumerate(channels):
                if sig is not None:
                    entry = (sig.origin_uuid, sig.name)

                    if getattr(sig, "color", None):
                        color = sig.color or serde.COLORS[index % serde.COLORS_COUNT]
                    else:
                        color = serde.COLORS[index % serde.COLORS_COUNT]

                    others.append(
                        SignalOnline(
                            name=sig.name,
                            conversion=sig.conversion,
                            entry=entry,
                            unit=sig.unit,
                            format=getattr(sig, "format", "phys"),
                            color=color,
                        )
                    )

                    sig.ranges = copy_ranges(sig.ranges)

                    self.channels.dataView.ranges[entry] = sig.ranges

        else:
            others = []
            for index, sig in enumerate(channels):
                if sig is not None:
                    sig.flags &= ~sig.Flags.computed
                    sig.computation = None
                    exists = getattr(sig, "exists", True)
                    ranges = sig.ranges
                    sig = PlotSignal(sig, index=index, allow_trim=False, allow_nans=True)
                    if sig.conversion:
                        sig.phys_samples = sig.conversion.convert(sig.raw_samples, as_bytes=True)
                    sig.entry = sig.origin_uuid, sig.group_index, sig.channel_index

                    others.append(
                        SignalOffline(
                            signal=sig,
                            exists=exists,
                        )
                    )

                    self.channels.dataView.ranges[sig.entry] = ranges

        self.channels.backend.update(others)
        self.update_timebase()

    def reset(self):
        self.channels.backend.reset()
        self.channels.dataView.double_clicked_enabled = True

    def set_values(self, values=None):
        selection = self.channels.dataView.selectedIndexes()
        self.channels.backend.set_values(values)

        selection_model = self.channels.dataView.selectionModel()
        for index in selection:
            selection_model.select(index, QtCore.QItemSelectionModel.SelectionFlag.Select)

    def to_config(self):
        channels = []

        pattern = self.pattern
        if not pattern:
            for signal in self.channels.backend.signals:
                channels.append(
                    {
                        "origin_uuid": str(signal.entry[0]),
                        "name": signal.name,
                        "ranges": self.channels.dataView.ranges[signal.entry],
                        "format": signal.format,
                        "color": signal.color,
                    }
                )

        config = {
            "format": "Physical",
            "mode": self.mode,
            "channels": channels,
            "pattern": pattern,
            "float_precision": self.float_precision.currentIndex() - 1,
            "header_sections_width": self.channels.columnHeader.all_columns_width(),
            "font_size": self.font().pointSize(),
            "columns_visibility": self.channels.columnHeader.columns_visibility(),
            "sorting": self.channels.columnHeader.sorting(),
            "header_and_controls_visible": self.controls.isVisible(),
        }

        return config

    def does_not_exist(self, entry, exists=False):
        self.channels.backend.does_not_exist(entry, exists)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.visible_entries_modified = True

    def reset_visible_entries(self, arg):
        self.visible_entries_modified = True

    def set_format(self, fmt):
        fmt = fmt.lower()
        if fmt in ("phys", "physical"):
            fmt_s = "Physical"
            fmt = "phys"
        elif fmt in ("bin", "binary"):
            fmt_s = "Bin"
            fmt = "bin"
        elif fmt == "hex":
            fmt_s = "Hex"
        elif fmt == "ascii":
            fmt_s = "Ascii"
        else:
            fmt_s = "Physical"
            fmt = "phys"

        self.channels.dataView.set_format(fmt)
        self._settings.setValue("numeric_format", fmt_s)
        self.channels.backend.data_changed()

    def set_float_precision(self, index):
        self._settings.setValue("numeric_float_precision", index - 1)
        self.channels.dataView.model().float_precision = index - 1
        self.channels.backend.data_changed()

    def visible_entries(self):
        visible = set()

        if self.channels.backend.sorted_column_index in (1, 2):
            visible = set(self.channels.backend.map)

        else:
            rect = self.channels.dataView.viewport().rect()

            top = self.channels.dataView.indexAt(rect.topLeft()).row()
            bottom = self.channels.dataView.indexAt(rect.bottomLeft()).row()
            if top == -1:
                pass
            elif bottom == -1:
                visible = set(self.channels.backend.map)
            else:
                for row in range(top, bottom + 1):
                    visible.add(self.channels.backend.signals[row].entry)

        self.visible_entries_modified = True

        return visible

    def _timestamp_changed(self, stamp):
        if not self._inhibit:
            self.set_timestamp(stamp, spinbox=True)

    def _timestamp_slider_changed(self, idx):
        if not self._inhibit:
            if not len(self.timebase):
                return

            self.set_timestamp(self.timebase[idx])

    def set_timestamp(self, stamp=None, emit=True, spinbox=False):
        if stamp is None:
            if self._timestamp is None:
                if len(self.timebase):
                    stamp = self.timebase[0]
                else:
                    return
            else:
                stamp = self._timestamp

        if not len(self.timebase):
            return

        idx = np.searchsorted(self.timebase, stamp, side="right") - 1

        new_stamp = self.timebase[idx]

        if spinbox:
            if new_stamp == self._timestamp and stamp > new_stamp:
                idx += 1
                new_stamp = self.timebase[idx]

        self._timestamp = new_stamp

        self.channels.backend.set_timestamp(new_stamp)

        self._inhibit = True
        self.timestamp_slider.setValue(idx)
        self.timestamp.setValue(new_stamp)
        self._inhibit = False

        if emit:
            self.timestamp_changed_signal.emit(self, new_stamp)

    def search_forward(self):
        if self.op.currentIndex() < 0 or not self.target.text().strip() or not self.pattern_match.text().strip():
            self.match.setText("invalid input values")
            return

        operator = self.op.currentText()

        if self.match_type.currentText() == "Wildcard":
            wildcard = f"{os.urandom(6).hex()}_WILDCARD_{os.urandom(6).hex()}"
            text = self.pattern_match.text().strip()
            pattern = text.replace("*", wildcard)
            pattern = re.escape(pattern)
            pattern = pattern.replace(wildcard, ".*")
        else:
            pattern = self.pattern_match.text().strip()

        if self.case_sensitivity.currentText() == "Case sensitive":
            pattern = re.compile(pattern)
        else:
            pattern = re.compile(f"(?i){pattern}")

        matches = [sig for sig in self.channels.backend.signals if pattern.fullmatch(sig.name)]

        mode = self.match_mode.currentText()

        if not matches:
            self.match.setText("the pattern does not match any channel name")
            return

        try:
            target = float(self.target.text().strip())
        except:
            self.match.setText("the target must a numeric value")
        else:
            if target.is_integer():
                target = int(target)

            start = self.timestamp.value()

            timestamp = None
            signal_name = ""
            for sig in matches:
                sig = sig.signal.cut(start=start)
                if mode == "Raw" or sig.conversion is None:
                    samples = sig.raw_samples
                else:
                    samples = sig.phys_samples

                op = getattr(samples, OPS[operator])
                try:
                    idx = np.argwhere(op(target)).flatten()
                    if len(idx):
                        if len(idx) == 1 or sig.timestamps[idx[0]] != start:
                            timestamp_ = sig.timestamps[idx[0]]
                        else:
                            timestamp_ = sig.timestamps[idx[1]]

                        if timestamp is None or timestamp_ < timestamp:
                            timestamp = timestamp_
                            signal_name = sig.name
                except:
                    continue

            if timestamp is not None:
                self.timestamp.setValue(timestamp)
                self.match.setText(f"condition found for {signal_name}")
            else:
                self.match.setText("condition not found")

    def search_backward(self):
        if self.op.currentIndex() < 0 or not self.target.text().strip() or not self.pattern_match.text().strip():
            self.match.setText("invalid input values")
            return

        operator = self.op.currentText()

        if self.match_type.currentText() == "Wildcard":
            wildcard = f"{os.urandom(6).hex()}_WILDCARD_{os.urandom(6).hex()}"
            text = self.pattern_match.text().strip()
            pattern = text.replace("*", wildcard)
            pattern = re.escape(pattern)
            pattern = pattern.replace(wildcard, ".*")
        else:
            pattern = self.pattern_match.text().strip()

        if self.case_sensitivity.currentText() == "Case sensitive":
            pattern = re.compile(pattern)
        else:
            pattern = re.compile(f"(?i){pattern}")

        matches = [sig for sig in self.channels.backend.signals if pattern.fullmatch(sig.name)]

        mode = self.match_mode.currentText()

        if not matches:
            self.match.setText("the pattern does not match any channel name")
            return

        try:
            target = float(self.target.text().strip())
        except:
            self.match.setText("the target must a numeric value")
        else:
            if target.is_integer():
                target = int(target)

            stop = self.timestamp.value()

            timestamp = None
            signal_name = ""
            for sig in matches:
                sig = sig.signal.cut(stop=stop)
                if mode == "raw values" or sig.conversion is None:
                    samples = sig.raw_samples[:-1]
                else:
                    samples = sig.phys_samples[:-1]

                op = getattr(samples, OPS[operator])
                try:
                    idx = np.argwhere(op(target)).flatten()
                    if len(idx):
                        if len(idx) == 1 or sig.timestamps[idx[-1]] != stop:
                            timestamp_ = sig.timestamps[idx[-1]]
                        else:
                            timestamp_ = sig.timestamps[idx[-2]]

                        if timestamp is None or timestamp_ > timestamp:
                            timestamp = timestamp_
                            signal_name = sig.name
                except:
                    continue

            if timestamp is not None:
                self.timestamp.setValue(timestamp)
                self.match.setText(f"condition found for {signal_name}")
            else:
                self.match.setText("condition not found")

    def color_same_origin_signals(self, origin_uuid="", color=""):
        self.backend.shift_same_origin_signals(origin_uuid=origin_uuid, color=color)

    def shift_same_origin_signals(self, origin_uuid="", delta=0.0):
        self.backend.shift_same_origin_signals(origin_uuid=origin_uuid, delta=delta)

    def update_missing_signals(self, uuids=()):
        self.channels.backend.signals = [
            signal for signal in self.channels.backend.signals if signal.origin_uuid in uuids
        ]

        self.channels.backend.data_changed()

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()

        if (
            key in (QtCore.Qt.Key.Key_H, QtCore.Qt.Key.Key_B, QtCore.Qt.Key.Key_P, QtCore.Qt.Key.Key_T)
            and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier
        ):
            event.accept()

            if key == QtCore.Qt.Key.Key_H:
                self.set_format("Hex")
            elif key == QtCore.Qt.Key.Key_B:
                self.set_format("Bin")
            elif key == QtCore.Qt.Key.Key_T:
                self.set_format("Ascii")
            else:
                self.set_format("Physical")

        elif (
            key
            in (
                QtCore.Qt.Key.Key_Left,
                QtCore.Qt.Key.Key_Right,
                QtCore.Qt.Key.Key_PageUp,
                QtCore.Qt.Key.Key_PageDown,
                QtCore.Qt.Key.Key_Home,
                QtCore.Qt.Key.Key_End,
            )
            and modifiers == QtCore.Qt.KeyboardModifier.NoModifier
            and self.mode == "offline"
        ):
            self.timestamp_slider.keyPressEvent(event)

        elif (
            key == QtCore.Qt.Key.Key_S
            and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier
            and self.mode == "offline"
        ):
            event.accept()
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
                self,
                "Save as measurement file",
                "",
                "MDF version 4 files (*.mf4)",
            )

            if file_name:
                signals = [offline_signal.signal for offline_signal in self.channels.dataView.backend.signals]
                if signals:
                    with mdf_module.MDF() as mdf:
                        groups = {}
                        for sig in signals:
                            id_ = id(sig.timestamps)
                            group_ = groups.setdefault(id_, [])
                            group_.append(sig)

                        for signals in groups.values():
                            sigs = []
                            for signal in signals:
                                if ":" in signal.name:
                                    sig = signal.copy()
                                    sig.name = sig.name.split(":")[-1].strip()
                                    sigs.append(sig)
                                else:
                                    sigs.append(signal)
                            mdf.append(sigs, common_timebase=True)
                        mdf.save(file_name, overwrite=True)

        elif key == QtCore.Qt.Key.Key_BracketLeft and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            event.accept()
            self.decrease_font()

        elif key == QtCore.Qt.Key.Key_BracketRight and modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            event.accept()
            self.increase_font()

        elif (
            key == QtCore.Qt.Key.Key_G
            and modifiers == QtCore.Qt.KeyboardModifier.ShiftModifier
            and self.mode == "offline"
        ):
            event.accept()

            value, ok = QtWidgets.QInputDialog.getDouble(
                self,
                "Go to time stamp",
                "Time stamp",
                value=self.timestamp_slider.value(),
                decimals=9,
            )

            if ok:
                self.set_timestamp(value)

        else:
            self.channels.dataView.keyPressEvent(event)

    def close(self):
        self.owner = None
        super().close()

    def decrease_font(self):
        font = self.font()
        size = font.pointSize()
        pos = bisect.bisect_left(FONT_SIZE, size) - 1
        pos = max(pos, 0)
        new_size = FONT_SIZE[pos]

        self.set_font_size(new_size)

    def increase_font(self):
        font = self.font()
        size = font.pointSize()
        pos = bisect.bisect_right(FONT_SIZE, size)
        if pos == len(FONT_SIZE):
            pos -= 1
        new_size = FONT_SIZE[pos]

        self.set_font_size(new_size)

    def set_font_size(self, size):
        self.hide()
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)
        self.show()
        self.channels.default_row_height = 12 + size
        self.channels.set_styles()

    def toggle_controls(self, event=None):
        if self.toggle_controls_btn.text() == "Show controls":
            self.toggle_controls_btn.setText("Hide controls")
            self.time_group.setHidden(False)
            self.search_group.setHidden(False)
            icon = QtGui.QIcon()
            icon.addPixmap(QtGui.QPixmap(":/up.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
            self.toggle_controls_btn.setIcon(icon)
        else:
            self.toggle_controls_btn.setText("Show controls")
            self.time_group.setHidden(True)
            self.search_group.setHidden(True)
            icon = QtGui.QIcon()
            icon.addPixmap(QtGui.QPixmap(":/down.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
            self.toggle_controls_btn.setIcon(icon)

    def update_timebase(self):
        if self.mode == "online":
            return

        self.timebase = self.channels.backend.timebase

        count = len(self.timebase)

        if count:
            min_, max_ = self.timebase[0], self.timebase[-1]
            self.timestamp_slider.setRange(0, count - 1)
            if count >= 2:
                self.timestamp.setSingleStep(0.5 * np.min(np.diff(self.timebase)))

        else:
            min_, max_ = 0.0, 0.0
            self.timestamp_slider.setRange(0, 0)

        self.timestamp.setRange(min_, max_)
        self.timestamp.setSingleStep(0.001)

        self.min_t.setText(f"{min_:.9f}s")
        self.max_t.setText(f"{max_:.9f}s")

        self.set_timestamp(emit=False)



from traceback import format_exc
import typing
from typing import Final

from canmatrix import Frame, Signal
import numpy as np
from numpy.typing import NDArray
from typing_extensions import Any, TypedDict

from . import v4_blocks as v4b
from . import v4_constants as v4c
from .conversion_utils import from_dict
from .utils import as_non_byte_sized_signed_int, MdfException

MAX_VALID_J1939: Final = {
    2: 1,
    4: 0xA,
    8: 0xFA,
    10: 0x3FA,
    12: 0xFAF,
    16: 0xFAFF,
    20: 0xFAFFF,
    24: 0xFAFFFF,
    28: 0xFAFFFFF,
    32: 0xFAFFFFFF,
    64: 0xFFFFFFFFFFFFFFFF,
}


def defined_j1939_bit_count(signal: Signal) -> int:
    size = typing.cast(int, signal.size)
    for defined_size in (2, 4, 8, 10, 12, 16, 20, 24, 28, 32, 64):
        if size <= defined_size:
            return defined_size
    return size


def apply_conversion(vals: NDArray[Any], signal: Signal, ignore_value2text_conversion: bool) -> NDArray[Any]:
    conv = get_conversion(signal)
    if conv and not (ignore_value2text_conversion and conv.conversion_type in v4c.CONVERSIONS_WITH_TEXTS):
        vals = conv.convert(vals)

    return vals


def extract_signal(
    signal: Signal,
    payload: NDArray[Any],
    raw: bool = False,
    ignore_value2text_conversion: bool = True,
    is_ISOTP: bool = False,
) -> NDArray[Any]:
    vals = payload

    big_endian = False if signal.is_little_endian else True
    signed = signal.is_signed
    is_float = signal.is_float

    start_bit = signal.get_startbit(bit_numbering=1)

    if big_endian:
        start_byte = start_bit // 8
        bit_count = signal.size

        pos = start_bit % 8 + 1

        over = bit_count % 8

        if pos >= over:
            bit_offset = (pos - over) % 8
        else:
            bit_offset = pos + 8 - over
    else:
        start_byte, bit_offset = divmod(start_bit, 8)

    bit_count = signal.size

    if is_float:
        if bit_offset:
            raise MdfException(f"Cannot extract float signal '{signal}' because it is not byte aligned")
        if bit_count not in (16, 32, 64):
            raise MdfException(f"Cannot extract float signal '{signal}' because it does not have a standard byte size")

    if big_endian:
        byte_pos = start_byte + 1
        start_pos = start_bit
        bits = bit_count

        while True:
            pos = start_pos % 8 + 1
            if pos < bits:
                byte_pos += 1
                bits -= pos
                start_pos = 7
            else:
                break

        if byte_pos > vals.shape[1]:
            raise MdfException(
                f'Could not extract signal "{signal.name}" with start '
                f"bit {start_bit} and bit count {signal.size} "
                f"from the payload with shape {vals.shape}"
            )
    else:
        if start_bit + bit_count > vals.shape[1] * 8:
            raise MdfException(
                f'Could not extract signal "{signal.name}" with start '
                f"bit {start_bit} and bit count {signal.size} "
                f"from the payload with shape {vals.shape}"
            )

    byte_size, r = divmod(bit_offset + bit_count, 8)
    if r:
        byte_size += 1

    if byte_size in (1, 2, 4, 8):
        extra_bytes = 0
    else:
        extra_bytes = 4 - (byte_size % 4)

    std_size = byte_size + extra_bytes

    # prepend or append extra bytes columns
    # to get a standard size number of bytes
    if is_ISOTP:  # Don't muck around with size of ISO-TP signals
        return vals

    if extra_bytes:
        if big_endian:
            vals = np.column_stack(
                [
                    vals[:, start_byte : start_byte + byte_size],
                    np.zeros(len(vals), dtype=f"<({extra_bytes},)u1"),
                ]
            )

            if std_size > 8:
                fmt = f"({std_size},)u1"
            elif is_float:
                fmt = f">f{std_size}"
            else:
                fmt = f">u{std_size}"

            try:
                vals = vals.view(fmt).ravel()
            except:
                vals = np.frombuffer(vals.tobytes(), dtype=fmt)

            if std_size <= 8 and not is_float:
                vals = vals >> (extra_bytes * 8 + bit_offset)
                vals &= (2**bit_count) - 1

        else:
            vals = np.column_stack(
                [
                    vals[:, start_byte : start_byte + byte_size],
                    np.zeros(len(vals), dtype=f"<({extra_bytes},)u1"),
                ]
            )

            if std_size > 8:
                fmt = f"({std_size},)u1"
            elif is_float:
                fmt = f"<f{std_size}"
            else:
                fmt = f"<u{std_size}"

            try:
                vals = vals.view(fmt).ravel()
            except:
                vals = np.frombuffer(vals.tobytes(), dtype=fmt)

            if std_size <= 8 and not is_float:
                vals = vals >> bit_offset
                vals &= (2**bit_count) - 1

    else:
        if big_endian:
            if std_size > 8:
                fmt = f"({std_size},)u1"
            elif is_float:
                fmt = f">f{std_size}"
            else:
                fmt = f">u{std_size}"

            try:
                vals = vals[:, start_byte : start_byte + byte_size].view(fmt).ravel()
            except:
                vals = np.frombuffer(
                    vals[:, start_byte : start_byte + byte_size].tobytes(),
                    dtype=fmt,
                )

            if std_size <= 8 and not is_float:
                vals = vals >> bit_offset
                vals &= (2**bit_count) - 1
        else:
            if std_size > 8:
                fmt = f"({std_size},)u1"
            elif is_float:
                fmt = f"<f{std_size}"
            else:
                fmt = f"<u{std_size}"

            try:
                vals = vals[:, start_byte : start_byte + byte_size].view(fmt).ravel()
            except:
                vals = np.frombuffer(
                    vals[:, start_byte : start_byte + byte_size].tobytes(),
                    dtype=fmt,
                )

            if std_size <= 8 and not is_float:
                vals = vals >> bit_offset
                vals &= (2**bit_count) - 1

    if signed and not is_float:
        if extra_bytes or bit_count not in (8, 16, 32, 64):
            vals = as_non_byte_sized_signed_int(vals, bit_count)
        else:
            vals = vals.view(f"i{std_size}")

    if not raw:
        vals = apply_conversion(vals, signal, ignore_value2text_conversion)

    return vals


def extract_can_signal(
    signal: Signal,
    payload: NDArray[Any],
    raw: bool = False,
    ignore_value2text_conversion: bool = True,
) -> NDArray[Any]:
    return extract_signal(signal, payload, raw, ignore_value2text_conversion)


def extract_lin_signal(
    signal: Signal,
    payload: NDArray[Any],
    raw: bool = False,
    ignore_value2text_conversion: bool = True,
) -> NDArray[Any]:
    return extract_signal(signal, payload, raw, ignore_value2text_conversion)


class ExtractedSignal(TypedDict):
    name: str
    comment: str
    unit: str
    samples: NDArray[Any]
    conversion: v4b.ChannelConversion | None
    t: NDArray[Any]
    invalidation_bits: NDArray[np.bool] | None


def merge_cantp(payload: NDArray[Any], ts: NDArray[Any]) -> tuple[NDArray[Any], NDArray[Any]]:
    """Merge sequences of ISO-TP coded CAN payloads, enabling > 8 byte frames."""
    INITIAL = 0x10
    CONSECUTIVE = 0x20
    merged = []
    t_out = []
    merging = np.array([], "uint8")
    for frame, t in zip(payload, ts, strict=False):
        if frame[0] & 0xF0 == INITIAL:
            expected_size = np.uint16(256) * (frame[0] & 0x0F) + frame[1]
            merging = np.array(frame[2:8], "uint8")
        if frame[0] & 0xF0 == CONSECUTIVE:
            merging = np.hstack((merging, frame[1:]))
            if len(merging) >= expected_size:
                merging = merging[:expected_size]
                merged.append(merging)
                t_out.append(t)  # Using t from final received part (as does Canoe, apparently)
    frames = np.vstack(merged) if len(merged) > 0 else np.array([], "uint8")
    return frames, np.array(t_out)


def extract_mux(
    payload: NDArray[Any],
    message: Frame,
    message_id: int | None,
    bus: int | None,
    t: NDArray[Any],
    muxer: str | None = None,
    muxer_values: NDArray[Any] | None = None,
    original_message_id: int | None = None,
    raw: bool = False,
    include_message_name: bool = False,
    ignore_value2text_conversion: bool = True,
    is_j1939: bool = False,
    is_extended: bool = False,
) -> dict[tuple[int | None, int | None, bool, int | None, str | None, int, int], dict[str, ExtractedSignal]]:
    """Extract multiplexed CAN signals from the raw payload.

    Parameters
    ----------
    payload : np.ndarray
        Raw CAN payload as numpy array.
    message : canmatrix.Frame
        CAN message description parsed by canmatrix.
    message_id : int
        Message id.
    bus : int
        Bus channel number.
    t : np.ndarray
        Timestamps for the raw payload.
    muxer : str, optional
        Name of the parent multiplexor signal.
    muxer_values : np.ndarray, optional
        Multiplexor signal values.
    original_message_id : int, optional
        Original message id.
    ignore_value2text_conversion : bool, default True
        Ignore value to text conversions.

        .. versionadded:: 5.23.0

    Returns
    -------
    extracted_signal : dict
        Each value in the dict is a list of signals that share the same
        multiplexors.
    """

    if muxer is None:
        if message.is_multiplexed:
            for sig in message:
                if sig.multiplex == "Multiplexor" and sig.muxer_for_signal is None:
                    multiplexor_name = sig.name
                    break
            for sig in message:
                if sig.multiplex not in (None, "Multiplexor"):
                    if sig.muxer_for_signal is None:
                        sig.muxer_for_signal = multiplexor_name
                    if not hasattr(sig, "mux_val_min"):
                        sig.mux_val_min = sig.mux_val_max = int(sig.multiplex)
                        sig.mux_val_grp.insert(0, (int(sig.multiplex), int(sig.multiplex)))

    extracted_signals: dict[
        tuple[int | None, int | None, bool, int | None, str | None, int, int], dict[str, ExtractedSignal]
    ] = {}

    # (Too?) simple check for ISO-TP CAN data - if it has flow control, we believe its ISO-TP
    is_ISOTP = "CanTpFcFrameId" in message.attributes
    if is_ISOTP:
        # print(f"  ISO-TP frame, for message {message_id}, merging CAN frames...")
        payload, t = merge_cantp(payload, t)
        # assert(len(payload) == len(t))
        # if len(payload) > 0:
        #    print(f"    message size post-merge: {payload.shape[1]}")
        # else:
        #    print(f"    no payload found to merge")

    if message.size == 0 or payload.shape[1] == 0:
        return extracted_signals

    elif message.size > payload.shape[1]:
        extra_bytes = message.size - payload.shape[1]
        payload = np.column_stack(
            [
                payload,
                np.full(len(payload), 0xFF, dtype=f"({extra_bytes},)u1"),
            ]
        )

    pairs: dict[tuple[int, int], list[Signal]] = {}
    for signal in message:
        if signal.muxer_for_signal == muxer:
            try:
                pair = signal.mux_val_min, signal.mux_val_max
            except:
                pair = tuple(signal.mux_val_grp[0]) if signal.mux_val_grp else (0, 0)
            pair_signals = pairs.setdefault(pair, [])
            pair_signals.append(signal)

    for pair, pair_signals in pairs.items():
        entry = bus, message_id, is_extended, original_message_id, muxer, *pair

        signals = extracted_signals.setdefault(entry, {})

        if muxer_values is not None:
            min_, max_ = pair
            idx = np.argwhere((min_ <= muxer_values) & (muxer_values <= max_)).ravel()
            payload_ = payload[idx]
            t_ = t[idx]
        else:
            t_ = t
            payload_ = payload

        for sig in pair_signals:
            samples = extract_signal(
                sig,
                payload_,
                ignore_value2text_conversion=ignore_value2text_conversion,
                raw=True,
                is_ISOTP=is_ISOTP,
            )
            if len(samples) == 0 and len(t_):
                continue

            if include_message_name:
                sig_name = f"{message.name}.{sig.name}"
            else:
                sig_name = sig.name

            try:
                scale_ranges = getattr(sig, "scale_ranges", None)
                if scale_ranges:
                    unit = scale_ranges[0]["unit"] or ""
                else:
                    unit = sig.unit or ""

                signals[sig_name] = {
                    "name": sig_name,
                    "comment": sig.comment or "",
                    "unit": unit,
                    "samples": samples if raw else apply_conversion(samples, sig, ignore_value2text_conversion),
                    "conversion": get_conversion(sig) if raw else None,
                    "t": t_,
                    "invalidation_bits": None,
                }

                if is_j1939:
                    signals[sig_name]["invalidation_bits"] = samples > MAX_VALID_J1939[defined_j1939_bit_count(sig)]

            except:
                print(format_exc())
                print(message, sig)
                print(samples, set(samples), samples.dtype, samples.shape)
                raise

            if sig.multiplex == "Multiplexor":
                extracted_signals.update(
                    extract_mux(
                        payload_,
                        message,
                        message_id,
                        bus,
                        t_,
                        muxer=sig.name,
                        muxer_values=samples,
                        original_message_id=original_message_id,
                        ignore_value2text_conversion=ignore_value2text_conversion,
                        raw=raw,
                        is_j1939=is_j1939,
                        is_extended=is_extended,
                    )
                )

    return extracted_signals


def get_conversion(signal: Signal) -> v4b.ChannelConversion:
    conv: v4b.ChannelConversionKwargs = {}

    a, b = float(signal.factor), float(signal.offset)

    scale_ranges = getattr(signal, "scale_ranges", None)
    if scale_ranges:
        for i, scale_info in enumerate(scale_ranges):
            conv[f"upper_{i}"] = scale_info["max"]  # type: ignore[literal-required]
            conv[f"lower_{i}"] = scale_info["min"]  # type: ignore[literal-required]
            conv[f"text_{i}"] = from_dict({"a": scale_info["factor"], "b": scale_info["offset"]})  # type: ignore[literal-required]

        for i, (val, text) in enumerate(signal.values.items(), len(scale_ranges)):
            conv[f"upper_{i}"] = val  # type: ignore[literal-required]
            conv[f"lower_{i}"] = val  # type: ignore[literal-required]
            conv[f"text_{i}"] = text  # type: ignore[literal-required]

        conv["default_addr"] = from_dict({"a": a, "b": b})

    elif signal.values:
        for i, (val, text) in enumerate(signal.values.items()):
            conv[f"upper_{i}"] = val  # type: ignore[literal-required]
            conv[f"lower_{i}"] = val  # type: ignore[literal-required]
            conv[f"text_{i}"] = text  # type: ignore[literal-required]

        conv["default_addr"] = from_dict({"a": a, "b": b})

    else:
        conv["a"] = a
        conv["b"] = b

    return from_dict(conv)



from abc import ABC
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
import logging
from pathlib import Path
from typing import Generic

import numpy as np
from numpy.typing import DTypeLike, NDArray
from typing_extensions import Any, Required, TypedDict, TypeVar

from . import v2_v3_blocks as v3b
from . import v4_blocks as v4b
from .types import DbcFileType, StrPath
from .utils import (
    ChannelsDB,
    DataBlockInfo,
    EMPTY_TUPLE,
    MdfException,
    SignalDataBlockInfo,
)

logger = logging.getLogger("asammdf")

__all__ = ["MDF_Common"]


class MdfKwargs(TypedDict, total=False):
    temporary_folder: StrPath | None
    raise_on_multiple_occurrences: bool
    use_display_names: bool
    fill_0_for_missing_computation_channels: bool
    remove_source_from_channel_names: bool
    password: str | None
    progress: Callable[[int, int], None] | Any
    callback: Callable[[int, int], None] | Any


class MdfCommonKwargs(MdfKwargs, total=False):
    original_name: Required[str | Path | None]
    __internal__: bool


_DG = TypeVar("_DG", v3b.DataGroup, v4b.DataGroup)
_CG = TypeVar("_CG", v3b.ChannelGroup, v4b.ChannelGroup)
_CN = TypeVar("_CN", v3b.Channel, v4b.Channel)
_CD = TypeVar("_CD", v3b.ChannelDependency, list[v4b.ChannelArrayBlock] | list[tuple[int, int]])
_ST = TypeVar("_ST", list[int], list[int] | list[tuple[int, int]] | np.dtype[Any])


class Group(Generic[_DG, _CG, _CN, _CD, _ST]):
    __slots__ = (
        "channel_dependencies",
        "channel_group",
        "channels",
        "data_blocks",
        "data_blocks_info_generator",
        "data_group",
        "data_location",
        "index",
        "read_split_count",
        "record",
        "record_size",
        "signal_data",
        "signal_types",
        "single_channel_dtype",
        "sorted",
        "string_dtypes",
        "trigger",
        "uses_ld",
        "uuid",
    )

    def __init__(self, data_group: _DG) -> None:
        self.data_group: _DG = data_group
        self.channel_group: _CG
        self.channels: list[_CN] = []
        self.channel_dependencies: list[_CD | None] = []
        self.signal_data: list[tuple[list[SignalDataBlockInfo], Iterator[SignalDataBlockInfo]] | None] = []
        self.record: list[tuple[np.dtype[Any], int, int, int] | None] | None = None
        self.record_size: dict[int, int] = {}
        self.trigger: v3b.TriggerBlock | None = None
        self.sorted: bool
        self.string_dtypes: list[np.dtype[np.bytes_]] = []
        self.data_blocks: list[DataBlockInfo] = []
        self.signal_types: _ST
        self.single_channel_dtype: DTypeLike | None = None
        self.uses_ld = False
        self.read_split_count = 0
        self.data_blocks_info_generator: Iterator[DataBlockInfo] = iter(EMPTY_TUPLE)
        self.uuid = ""
        self.data_location: int
        self.index = 0

    def set_blocks_info(self, info: list[DataBlockInfo]) -> None:
        self.data_blocks = info

    def __contains__(self, item: str) -> bool:
        return hasattr(self, item)

    def clear(self) -> None:
        self.data_blocks.clear()
        self.channels.clear()
        self.channel_dependencies.clear()
        self.signal_data.clear()
        self.data_blocks_info_generator = iter(())

    def get_data_blocks(self) -> Iterator[DataBlockInfo]:
        yield from self.data_blocks

        while True:
            try:
                info = next(self.data_blocks_info_generator)
                self.data_blocks.append(info)
                yield info
            except StopIteration:
                break

    def get_signal_data_blocks(self, index: int) -> Iterator[SignalDataBlockInfo]:
        signal_data = self.signal_data[index]
        if signal_data is not None:
            signal_data_blocks, signal_generator = signal_data
            yield from signal_data_blocks

            while True:
                try:
                    info = next(signal_generator)
                    signal_data_blocks.append(info)
                    yield info
                except StopIteration:
                    break

    def load_all_data_blocks(self) -> None:
        for _ in self.get_data_blocks():
            continue


GroupV3 = Group[v3b.DataGroup, v3b.ChannelGroup, v3b.Channel, v3b.ChannelDependency, list[int]]
GroupV4 = Group[
    v4b.DataGroup,
    v4b.ChannelGroup,
    v4b.Channel,
    list[v4b.ChannelArrayBlock] | list[tuple[int, int]],
    list[int] | list[tuple[int, int]] | np.dtype[Any],
]

_Group = TypeVar("_Group", GroupV3, GroupV4)


class MDF_Common(ABC, Generic[_Group]):
    """Common methods for MDF objects."""

    def __init__(self, raise_on_multiple_occurrences: bool) -> None:
        self.groups: list[_Group] = []
        self.channels_db = ChannelsDB()
        self._raise_on_multiple_occurrences = raise_on_multiple_occurrences

    def _set_temporary_master(self, master: NDArray[Any] | None) -> None:
        self._master = master

    # @lru_cache(maxsize=1024)
    def _validate_channel_selection(
        self,
        name: str | None = None,
        group: int | None = None,
        index: int | None = None,
    ) -> tuple[int, int]:
        """Validate channel selection.

        The channel can be specified in two ways:

        * Using the first positional argument `name`.

          * If there are multiple occurrences for this channel, then the `group`
            and `index` arguments can be used to select a specific group.
          * If there are multiple occurrences for this channel and either the
            `group` or `index` arguments is None, then a warning is issued.

        * Using the group number (keyword argument `group`) and the channel
          number (keyword argument `index`). Use `info` method for group and
          channel numbers.

        Parameters
        ----------
        name : str, optional
            Name of channel.
        group : int, optional
            0-based group index.
        index : int, optional
            0-based channel index.

        Returns
        -------
        group_index, channel_index : (int, int)
            Selected channel's group and channel index.

        Raises
        ------
        MdfException
            * if the channel name is not found
            * if the group index is out of range
            * if the channel index is out of range
            * if there are multiple channel occurrences in the file and the
              arguments `name`, `group`, `index` are ambiguous. This behaviour
              can be turned off by setting `raise_on_multiple_occurrences` to
              False.
        """

        if name is None:
            if group is None or index is None:
                message = 'Invalid arguments for channel selection: must give "name" or, "group" and "index"'
                raise MdfException(message)
            else:
                gp_nr, ch_nr = group, index
                if ch_nr >= 0:
                    try:
                        grp = self.groups[gp_nr]
                    except IndexError:
                        raise MdfException("Group index out of range") from None

                    try:
                        grp.channels[ch_nr]
                    except IndexError:
                        raise MdfException(f"Channel index out of range: {(name, group, index)}") from None
        else:
            if name not in self.channels_db:
                raise MdfException(f'Channel "{name}" not found')
            else:
                if group is None:
                    entries = self.channels_db[name]
                    if len(entries) > 1:
                        if self._raise_on_multiple_occurrences:
                            message = (
                                f'Multiple occurrences for channel "{name}": {entries}. '
                                'Provide both "group" and "index" arguments'
                                " to select another data group"
                            )
                            logger.exception(message)
                            raise MdfException(message)
                        else:
                            message = (
                                f'Multiple occurrences for channel "{name}": {entries}. '
                                "Returning the first occurrence since the MDF object was "
                                "configured to not raise an exception in this case."
                            )
                            logger.warning(message)
                            gp_nr, ch_nr = entries[0]
                    else:
                        gp_nr, ch_nr = entries[0]

                else:
                    if index is not None and index < 0:
                        gp_nr = group
                        ch_nr = index
                    else:
                        if index is None:
                            entries = tuple((gp_nr, ch_nr) for gp_nr, ch_nr in self.channels_db[name] if gp_nr == group)
                            count = len(entries)

                            if count == 1:
                                gp_nr, ch_nr = entries[0]

                            elif count == 0:
                                message = f'Channel "{name}" not found in group {group}'
                                raise MdfException(message)

                            else:
                                if self._raise_on_multiple_occurrences:
                                    message = (
                                        f'Multiple occurrences for channel "{name}": {entries}. '
                                        'Provide both "group" and "index" arguments'
                                        " to select another data group"
                                    )
                                    logger.exception(message)
                                    raise MdfException(message)
                                else:
                                    message = (
                                        f'Multiple occurrences for channel "{name}": {entries}. '
                                        "Returning the first occurrence since the MDF object was "
                                        "configured to not raise an exception in this case."
                                    )
                                    logger.warning(message)
                                    gp_nr, ch_nr = entries[0]
                        else:
                            if (group, index) in self.channels_db[name]:
                                ch_nr = index
                                gp_nr = group
                            else:
                                message = f'Channel "{name}" not found in group {group} at index {index}'
                                raise MdfException(message)

        return gp_nr, ch_nr


class _BusInfo(TypedDict):
    dbc_files: Iterable[DbcFileType]
    unknown_id_count: int


class CanInfo(_BusInfo):
    total_unique_ids: set[tuple[int, bool]]
    not_found_ids: defaultdict[StrPath, list[tuple[tuple[int, bool] | int, str]]]
    found_ids: defaultdict[StrPath, set[tuple[tuple[int, int, bool], str]]]
    unknown_ids: set[int | tuple[int, bool]]
    max_flags: list[list[list[bool]]]


class LinInfo(_BusInfo):
    total_unique_ids: set[tuple[int, ...]]
    not_found_ids: defaultdict[StrPath, list[tuple[int, str]]]
    found_ids: defaultdict[StrPath, set[tuple[tuple[int, bool, bool], str]]]
    unknown_ids: set[int]


class LastCallInfo(TypedDict, total=False):
    CAN: CanInfo
    LIN: LinInfo



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'channel_group_info_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGroupBox, QHBoxLayout,
    QLabel, QListWidget, QListWidgetItem, QPlainTextEdit,
    QScrollArea, QScrollBar, QSizePolicy, QSpacerItem,
    QSplitter, QTabWidget, QVBoxLayout, QWidget)

class Ui_ChannelGroupInfo(object):
    def setupUi(self, ChannelGroupInfo):
        if not ChannelGroupInfo.objectName():
            ChannelGroupInfo.setObjectName(u"ChannelGroupInfo")
        ChannelGroupInfo.resize(926, 558)
        self.verticalLayout_2 = QVBoxLayout(ChannelGroupInfo)
        self.verticalLayout_2.setSpacing(1)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.tabs = QTabWidget(ChannelGroupInfo)
        self.tabs.setObjectName(u"tabs")
        font = QFont()
        font.setFamilies([u"Consolas"])
        font.setPointSize(8)
        self.tabs.setFont(font)
        self.ta_widet = QWidget()
        self.ta_widet.setObjectName(u"ta_widet")
        self.horizontalLayout = QHBoxLayout(self.ta_widet)
        self.horizontalLayout.setSpacing(1)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(1, 1, 1, 1)
        self.groupBox = QGroupBox(self.ta_widet)
        self.groupBox.setObjectName(u"groupBox")
        self.groupBox.setFlat(False)
        self.verticalLayout_4 = QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setSpacing(1)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(1, 1, 1, 1)
        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 443, 497))
        self.verticalLayout_3 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_3.setSpacing(1)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(1, 1, 1, 1)
        self.channel_group_label = QLabel(self.scrollAreaWidgetContents)
        self.channel_group_label.setObjectName(u"channel_group_label")
        self.channel_group_label.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)

        self.verticalLayout_3.addWidget(self.channel_group_label)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_3.addItem(self.verticalSpacer_3)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout_4.addWidget(self.scrollArea)


        self.horizontalLayout.addWidget(self.groupBox)

        self.groupBox_3 = QGroupBox(self.ta_widet)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.verticalLayout_5 = QVBoxLayout(self.groupBox_3)
        self.verticalLayout_5.setSpacing(1)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(1, 1, 1, 1)
        self.scrollArea_3 = QScrollArea(self.groupBox_3)
        self.scrollArea_3.setObjectName(u"scrollArea_3")
        self.scrollArea_3.setWidgetResizable(True)
        self.scrollAreaWidgetContents_3 = QWidget()
        self.scrollAreaWidgetContents_3.setObjectName(u"scrollAreaWidgetContents_3")
        self.scrollAreaWidgetContents_3.setGeometry(QRect(0, 0, 442, 497))
        self.verticalLayout = QVBoxLayout(self.scrollAreaWidgetContents_3)
        self.verticalLayout.setSpacing(1)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.source_label = QLabel(self.scrollAreaWidgetContents_3)
        self.source_label.setObjectName(u"source_label")
        self.source_label.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)

        self.verticalLayout.addWidget(self.source_label)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.scrollArea_3.setWidget(self.scrollAreaWidgetContents_3)

        self.verticalLayout_5.addWidget(self.scrollArea_3)


        self.horizontalLayout.addWidget(self.groupBox_3)

        self.tabs.addTab(self.ta_widet, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.horizontalLayout_3 = QHBoxLayout(self.tab_2)
        self.horizontalLayout_3.setSpacing(1)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(1, 1, 1, 1)
        self.splitter = QSplitter(self.tab_2)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout_6 = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_6.setSpacing(1)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.wrap = QCheckBox(self.layoutWidget)
        self.wrap.setObjectName(u"wrap")
        self.wrap.setChecked(True)

        self.verticalLayout_6.addWidget(self.wrap)

        self.channels = QListWidget(self.layoutWidget)
        self.channels.setObjectName(u"channels")

        self.verticalLayout_6.addWidget(self.channels)

        self.splitter.addWidget(self.layoutWidget)
        self.display = QPlainTextEdit(self.splitter)
        self.display.setObjectName(u"display")
        self.display.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.display.setUndoRedoEnabled(False)
        self.display.setLineWrapMode(QPlainTextEdit.LineWrapMode.WidgetWidth)
        self.display.setReadOnly(True)
        self.splitter.addWidget(self.display)

        self.horizontalLayout_3.addWidget(self.splitter)

        self.scroll = QScrollBar(self.tab_2)
        self.scroll.setObjectName(u"scroll")
        self.scroll.setMaximum(9999)
        self.scroll.setOrientation(Qt.Orientation.Vertical)

        self.horizontalLayout_3.addWidget(self.scroll)

        self.tabs.addTab(self.tab_2, "")

        self.verticalLayout_2.addWidget(self.tabs)


        self.retranslateUi(ChannelGroupInfo)

        self.tabs.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(ChannelGroupInfo)
    # setupUi

    def retranslateUi(self, ChannelGroupInfo):
        ChannelGroupInfo.setWindowTitle(QCoreApplication.translate("ChannelGroupInfo", u"Form", None))
        self.groupBox.setTitle(QCoreApplication.translate("ChannelGroupInfo", u"Channel Group", None))
        self.channel_group_label.setText(QCoreApplication.translate("ChannelGroupInfo", u"Channel group metadata", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("ChannelGroupInfo", u"Source", None))
        self.source_label.setText(QCoreApplication.translate("ChannelGroupInfo", u"No source", None))
        self.tabs.setTabText(self.tabs.indexOf(self.ta_widet), QCoreApplication.translate("ChannelGroupInfo", u"Channel group CGBLOCK", None))
        self.wrap.setText(QCoreApplication.translate("ChannelGroupInfo", u"Text wrap", None))
        self.tabs.setTabText(self.tabs.indexOf(self.tab_2), QCoreApplication.translate("ChannelGroupInfo", u"Raw bytes", None))
    # retranslateUi
